# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
A module that provides functions for "aligning" images: specifically, it
provides functions for computing corrections to image ``WCS`` so that
image catalogs "align" to the reference catalog *on the sky*.

:Authors: Mihai Cara

:License: :doc:`../LICENSE`

"""
# STDLIB
import logging
from datetime import datetime
import collections

# THIRD PARTY
import numpy as np
import astropy

# LOCAL
from . wcsimage import RefCatalog, WCSImageCatalog, WCSGroupCatalog
from . tpwcs import TPWCS
from . matchutils import TPMatch
from . linearfit import SUPPORTED_FITGEOM_MODES, _SUPPORTED_FITGEOM_EN_STR

from . import __version__

__author__ = 'Mihai Cara'

__all__ = ['fit_wcs', 'align_wcs']


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def fit_wcs(refcat, imcat, tpwcs, ref_tpwcs=None, fitgeom='general', nclip=3,
            sigma=(3.0, 'rmse')):
    """ "Tweak" **a single** image's ``WCS`` by fitting image catalog to a
    reference catalog. This is a simplified version of `align_wcs` that does
    not perform matching and is limited to the fitting part.

    .. note::
        Both reference and image catalogs must have been **matched**
        *prior to* calling ``fit_wcs()``. This means that the lengths of both
        ``refcat`` and ``imcat`` catalogs must be equal *and* that coordinates
        with the same indices in both catalogs correspond to the same source.

    .. warning::
        If ``tpwcs.meta`` dictionary contains ``'catalog'`` keyword,
        it will be ignored.

    Parameters
    ----------
    refcat: astropy.table.Table
        A reference source catalog. The catalog must contain ``'RA'`` and
        ``'DEC'`` columns which indicate reference source world
        coordinates (in degrees). An optional column in the catalog is
        the ``'weight'`` column, which when present, will be used in fitting.
        See ``Notes`` section for further details.

    imcat: astropy.table.Table
        Source catalog associated with an image whose WCS needs to be aligned
        by fitting a linear transformation to ``imcat`` source positions so as
        to align them to the same sources from the ``refcat`` catalog.
        Must contain ``'x'`` and ``'y'`` columns which indicate source
        coordinates (in pixels) in the associated image. An optional column in
        the catalog is the ``'weight'`` column, which when present, will be
        used in fitting. See ``Notes`` section for further details.

    tpwcs: TPWCS
        A ``WCS`` associated with the image from which the catalog was derived.
        This ``TPWCS``-subclassed WCS corrector object must also define
        a tangent plane that will be used for fitting the two catalogs'
        sources and in which WCS corrections will be applied.

    ref_tpwcs: TPWCS, None, optional
        A reference WCS of the type ``TPWCS`` that provides the tangent
        plane in which matching will be performed and corrections will be
        defined. When not provided (i.e., set to `None`), reference tangent
        plane will be the same as defined by ``tpwcs`` argument.

    fitgeom: {'shift', 'rshift', 'rscale', 'general'}, optional
        The fitting geometry to be used in fitting the matched object lists.
        This parameter is used in fitting the offsets, rotations and/or scale
        changes from the matched object lists. The 'general' fit geometry
        allows for independent scale and rotation for each axis.

    nclip: int, None, optional
        Number (a non-negative integer) of clipping iterations in fit.
        Clipping will be turned off if ``nclip`` is either `None` or 0.

    sigma: float, tuple of the form (float, str), optional
        When a tuple is provided, first value (a positive number)
        indicates the number of "fit error estimates" to use for clipping.
        The second value (a string) indicates the statistic to be
        used for "fit error estimate". Currently the following values are
        supported: ``'rmse'``, ``'mae'``, and ``'std'``
        - see `~tweakwcs.linearfit.iter_linear_fit` for more details.

        When ``sigma`` is a single number, it must be a positive number and
        the default error estimate ``'rmse'`` is assumed.

        This parameter is ignored when ``nclip`` is either `None` or 0.

    Returns
    -------
    twwcs: TPWCS
        "Tweaked" (aligned) ``WCS`` that contains tangent-plane corrections
        so that reference and image catalog sources better align in the tangent
        plane and therefore on the sky as well.

    Notes
    -----
    When fitting image sources to reference catalog sources, we can specify
    which sources have higher weights. This can be done by assigning a "weight"
    to each source by specifying these values in the optional ``'weight'``
    column of either the reference catalog, image catalog, or both.

    When weights are not provided, all sources are weighed equally. When
    only either image or reference catalog weights are provided, the sources
    will be weighted with the specified weights. When *both* image *and*
    reference catalogs specify weights for the same sources, the two weights
    will be combined into a single weight as:

    .. math::
        1/w = 1/w_i + 1/w_r

    .. warning::
        Keep in mind that when a group catalog is created from individual
        catalogs, weights of the group catalog are created by
        *concatenating* weights of individual catalogs. Therefore,
        for the weighting of groups of catalogs to work correctly,
        the weights of individual catalogs should be scaled in such a way
        that when individual catalogs are combined into a single
        "group catalog", weights preserve their relative values.

        For example, let's say a group is formed from two individual
        catalogs. Let's say first catalog contains four sources with equal
        weights ``[1,1,1,1]`` and the second catalog contains two sources
        with weights ``[1,1]`` then the group's catalogs sources will
        also have equal weights ``[1,1,1,1,1,1]``. However, if each
        individual catalog's weights were normalized such that sum of
        all weights is 1, then group's sources will be weighed unequally:
        ``[0.25,0.25,0.25,0.25,0.5,0.5]``.

    Upon **successful** completion, this function will set the ``'fit_info'``
    key value of the ``meta`` attribute of the returned ``TPWCS`` object.
    ``'fit_info'`` is a dictionary with the following items:

        * **'shift'**: A ``numpy.ndarray`` with two components of the
          computed shift.

        * **'matrix'**: A ``2x2`` ``numpy.ndarray`` with the computed
          generalized rotation matrix.

        * **'proper_rot'**: Rotation angle (degree) as if the rotation is
          proper.

        * **'rot'**: A tuple of ``(rotx, roty)`` - the rotation angles with
          regard to the ``X`` and ``Y`` axes.

        * **'<rot>'**: *Arithmetic mean* of the angles of rotation around
          ``X`` and ``Y`` axes.

        * **'scale'**: A tuple of ``(sx, sy)`` - scale change in the direction
          of the ``X`` and ``Y`` axes.

        * **'<scale>'**: *Geometric mean* of scales ``sx`` and ``sy``.

        * **'skew'**: Computed skew.

        * **'proper'**: a boolean indicating whether the rotation is proper.

        * **'fitgeom'**: Fit geometry (allowed transformations) used for
          fitting data (to minimize residuals). This is copy of the input
          argument ``fitgeom``.

        * **'center'**: Center of rotation in the *tangent plane* of the
          computed linear transformations.

        * **'fitmask'**: A boolean array indicating which source positions
          where used for fitting (`True`) and which were clipped out
          (`False`). **NOTE:** For weighted fits, positions with zero
          weights are automatically excluded from the fits.

        * **'eff_nclip'**: Effective number of clipping iterations

        * **'rmse'**: fit Root-Mean-Square Error in *tangent plane*
          coordinates of corrected image source positions from reference
          source positions.

        * **'mae'**: fit Mean Absolute Error in *tangent plane*
          coordinates of corrected image source positions from reference
          source positions.

        * **'std'**: Norm of the standard deviation of the residuals
          in *tangent plane* along each axis.

        * **'resids'**: An array of residuals of the fit in the
          *tangent plane*.

          **NOTE:** Only the residuals for the "valid" points are reported
          here. Therefore the length of this array may be smaller than the
          length of input arrays of positions.

        * **'fit_RA'**: first (corrected) world coordinate of input source
          positions used in fitting.

        * **'fit_DEC'**: second (corrected) world coordinate of input
          source positions used in fitting.

        * **'status'**: Alignment status. Currently two possible status are
          possible ``'SUCCESS'`` or ``'FAILED: reason for failure'``.
          When alignment failed, the reason for failure is provided after
          alignment status.

    """
    function_name = fit_wcs.__name__

    # Time it
    runtime_begin = datetime.now()

    log.info(" ")
    log.info("***** {:s}.{:s}() started on {}"
             .format(__name__, function_name, runtime_begin))
    log.info("      Version {}".format(__version__))
    log.info(" ")

    try:
        # Attempt to set initial status to FAILED.
        tpwcs.meta['fit_info'] = {'status': 'FAILED: Unknown error'}
    except Exception:
        raise AttributeError("Unable to set/modify tpwcs.meta attribute.")

    # check fitgeom:
    fitgeom = fitgeom.lower()
    if fitgeom not in SUPPORTED_FITGEOM_MODES:
        raise ValueError(
            f"Unsupported 'fitgeom'. Valid values are: {_SUPPORTED_FITGEOM_EN_STR:s}"
        )

    wimcat = WCSImageCatalog(imcat, tpwcs,
                             name=imcat.meta.get('name', 'Unnamed'))
    wgcat = WCSGroupCatalog(wimcat, name=imcat.meta.get('name', 'Unnamed'))
    wrefcat = RefCatalog(refcat, name=imcat.meta.get('name', 'Unnamed'))

    succes = wgcat.align_to_ref(
        refcat=wrefcat,
        ref_tpwcs=ref_tpwcs,
        match=None,
        minobj=None,
        fitgeom=fitgeom,
        nclip=nclip,
        sigma=sigma
    )

    tpwcs.meta['fit_info'] = wimcat.fit_info
    if not succes:
        log.warning("Failed to align catalog '{}'.".format(wgcat.name))

    # log running time:
    runtime_end = datetime.now()
    log.info(" ")
    log.info("***** {:s}.{:s}() ended on {}"
             .format(__name__, function_name, runtime_end))
    log.info("***** {:s}.{:s}() TOTAL RUN TIME: {}"
             .format(__name__, function_name, runtime_end - runtime_begin))
    log.info(" ")

    return wgcat[0].tpwcs


def align_wcs(wcscat, refcat=None, ref_tpwcs=None, enforce_user_order=True,
              expand_refcat=False, minobj=None, match=TPMatch(),
              fitgeom='general', nclip=3, sigma=(3.0, 'rmse')):
    r"""
    Align (groups of) image catalogs by adjusting the parameters of their
    WCS based on fits between matched sources in these catalogs and a reference
    catalog which may be automatically created from one of the input ``wcscat``
    catalogs.

    .. warning::
        This function modifies the ``wcs`` attribute of each item
        in the input ``wcscat`` list!

    Upon completion, this function will add a field ``'fit_info'``
    to the ``meta`` attribute of the input WCS correctors (except of the one
    chosen as a reference catalog when ``refcat`` is `None`) containing
    a dictionary describing matching and fit results. For a description
    of the items in this dictionary, see
    :meth:`tweakwcs.wcsimage.WCSGroupCatalog.align_to_ref`. In addition to the
    status set by :meth:`~tweakwcs.wcsimage.WCSGroupCatalog.align_to_ref`,
    this function may set status to ``'REFERENCE'`` for an input image used
    as a reference image when a reference catalog is not provided.
    In this case no other fields in the ``'fit_info'`` will be present
    because a reference image is not being aligned. When alignment failed,
    the reason for failure is provided after alignment status.

    .. warning::

        Unless status in ``'fit_info'`` is ``'SUCCESS'``, there is no
        guarantee that other fields in ``'fit_info'`` are present or
        valid. Therefore, it is advisable verify that status is ``'SUCCESS'``
        before attempting to access other items, for example:

        >>> fit_info = wcscat[0].meta.get('fit_info')  # noqa
        >>> if fit_info['status'] == 'SUCCESS':
        ...     print("shifts: [{}, {}]".format(*fit_info['shift']))
        ... else:
        ...     print("tweak info not available for this image")

    Parameters
    ----------
    wcscat: tweakwcs.tpwcs.TPWCS, list of tweakwcs.tpwcs.TPWCS
        A list of all `~tweakwcs.tpwcs.TPWCS`-derived WCS correctors whose
        ``meta`` dictionary **must** contain ``'catalog'``
        item with a non-empty table value of type `astropy.table.Table`.
        This catalog must contain ``'x'`` and ``'y'`` columns which indicate
        source coordinates (in pixels) in the associated image. An optional
        column in the catalog is the ``'weight'`` column, which when present,
        will be used in fitting. See ``Notes`` section for further details.
        In addition to ``'catalog'``, the following items in the ``meta``
        dictionary are recognized/supported: ``'name'`` and ``'group_id'``.
        ``'name'`` is catalog's name and it used to identify catalog during
        logging. If ``'name'`` value is `None` or not present at all in the
        ``meta`` of a catalog, the name of that catalog will reported as
        ``'Unknown'``. Group ID that may be used for identifying catalogs
        that need to be aligned together. ``group_id`` must be hashable.
        If ``'group_id'`` is `None` or not provided, each input WCS/catalog
        will be aligned individually.

        .. note::
            Upon completion this function will add ``'fit_info'``
            item (a dictionary) to input object's ``meta`` dictionary.
            See **Notes** section for more details.

        .. warning::
            This function modifies the WCS of ``TPWCS`` objects by calling
            their :py:meth:`~tweakwcs.tpwcs.TPWCS.set_correction` method.

    refcat: astropy.table.Table, optional
        A reference source catalog. The catalog must contain ``'RA'`` and
        ``'DEC'`` columns which indicate reference source world
        coordinates (in degrees). An optional column in the catalog is
        the ``'weight'`` column, which when present, will be used in fitting.
        See ``Notes`` section for further details.

    ref_tpwcs: TPWCS, None, optional
        A reference WCS of the type ``TPWCS`` that provides the tangent
        plane in which matching will be performed and corrections will be
        defined. When not provided (i.e., set to `None`), reference tangent
        plane will be defined from the first ``TPWCS`` object
        *in the re-ordered* (if ``enforce_user_order`` was
        set to `True`) input list ``wcscat``.

    enforce_user_order: bool, optional
        Specifies whether images should be aligned in the order specified in
        the `file` input parameter or `align` should optimize the order
        of alignment by intersection area of the images. Default value (`True`)
        will align images in the user specified order, except when some images
        cannot be aligned in which case `align` will optimize the image
        alignment order. Alignment order optimization is available *only*
        when ``expand_refcat`` is `True`.

    expand_refcat: bool, optional
        Specifies whether to add new sources from just matched images to
        the reference catalog to allow next image to be matched against an
        expanded reference catalog. By delault, the reference catalog is not
        being expanded.

        If ``refcat`` is not `None` and contains an ``'id'`` column, then
        sources being added to the reference catalog will be assigned
        consecutive IDs that continue maximum ID in the ``refcat``.

        If one desires to uniquely associate source in the expanded catalog
        to their original catalogs, it is recommended that one assign unique
        IDs to all sources in all input catalogs **and** in the reference
        catalog in a separate column such as ``'uuid'``.

    minobj: int, None, optional
        Minimum number of identified objects from each input image to use
        in matching objects from other images. If the default `None` value is
        used then `align` will automatically deternmine the minimum number
        of sources from the value of the ``fitgeom`` parameter.

    match: MatchCatalogs, function, None, optional
        A callable that takes two arguments: a reference catalog and an
        image catalog. Both catalogs will have columns ``'TPx'`` and
        ``'TPy'`` that represent the source coordinates in some common
        (to both catalogs) coordinate system.

    fitgeom: {'shift', 'rshift', 'rscale', 'general'}, optional
        The fitting geometry to be used in fitting the matched object lists.
        This parameter is used in fitting the offsets, rotations and/or scale
        changes from the matched object lists. The 'general' fit geometry
        allows for independent scale and rotation for each axis.

    nclip: int, None, optional
        Number (a non-negative integer) of clipping iterations in fit.
        Clipping will be turned off if ``nclip`` is either `None` or 0.

    sigma: float, tuple of the form (float, str), optional
        When a tuple is provided, first value (a positive number)
        indicates the number of "fit error estimates" to use for clipping.
        The second value (a string) indicates the statistic to be
        used for "fit error estimate". Currently the following values are
        supported: ``'rmse'``, ``'mae'``, and ``'std'``
        - see `~tweakwcs.linearfit.iter_linear_fit` for more details.

        When ``sigma`` is a single number, it must be a positive number and
        the default error estimate ``'rmse'`` is assumed.

        This parameter is ignored when ``nclip`` is either `None` or 0.

    Returns
    -------
    eff_refcat: astropy.table.Table
        Effective reference catalog used for aligning all images. Depending
        on the values of the input parameters ``refcat``,
        ``enforce_user_order``, and ``expand_refcat``, effective
        reference catalog may be one of the input image catalogs, the original
        ``refcat`` catalog, an expanded ``refcat`` with a combination of
        source positions from all input images.

    Notes
    -----
    **1. Weights:**

    When fitting image sources to reference catalog sources, we can specify
    which sources have higher weights. This can be done by assigning a "weight"
    to each source by specifying these values in the optional ``'weight'``
    column of either the reference catalog, image catalog, or both.

    When weights are not provided, all sources are weighed equally. When
    only either image or reference catalog weights are provided, the sources
    will be weighted with the specified weights. When *both* image *and*
    reference catalogs specify weights for the same sources, the two weights
    will be combined into a single weight as:

    .. math::
        1/w = 1/w_i + 1/w_r

    .. warning::
        Keep in mind that when a group catalog is created from individual
        catalogs, weights of the group catalog are created by
        *concatenating* weights of individual catalogs. Therefore,
        for the weighting of groups of catalogs to work correctly,
        the weights of individual catalogs should be scaled in such a way
        that when individual catalogs are combined into a single
        "group catalog", weights preserve their relative values.

        For example, let's say a group is formed from two individual
        catalogs. Let's say first catalog contains four sources with equal
        weights ``[1,1,1,1]`` and the second catalog contains two sources
        with weights ``[1,1]`` then the group's catalogs sources will
        also have equal weights ``[1,1,1,1,1,1]``. However, if each
        individual catalog's weights were normalized such that sum of
        all weights is 1, then group's sources will be weighed unequally:
        ``[0.25,0.25,0.25,0.25,0.5,0.5]``.

    .. warning::
        When image catalogs contain optional ``'weight'`` column, then
        all image catalogs in a group must contain this column.

    **2.** ``'fit_info'``:

    Upon completion, this function will add ``'fit_info'``
    item (itself a dictionary) to input object's ``meta`` dictionary.
    If input objects are `~tweakwcs.tpwcs.TPWCS` WCS correctors, then
    ``TPWCS.meta['fit_info']`` will be set to a dictionary containing
    fit information.

    .. note::
        For `~tweakwcs.tpwcs.TPWCS` that are aligned in a group,
        the ``'matrix'`` and ``'shift'`` items in the ``'fit_info'``
        dictionary may differ from
        the values of the same items in ``TPWCS.meta`` dictionary. This is
        normal since WCS corrections (in `~tweakwcs.tpwcs.TPWCS`) are applied
        in the image's WCS plane while fit may be performed in a slightly
        different tangent plane.

    """
    function_name = align_wcs.__name__

    # Time it
    runtime_begin = datetime.now()

    log.info(" ")
    log.info("***** {:s}.{:s}() started on {}"
             .format(__name__, function_name, runtime_begin))
    log.info("      Version {}".format(__version__))
    log.info(" ")

    # Check that type of `wcscat` is correct and set initial status to FAILED:
    if isinstance(wcscat, TPWCS):
        wcscat = [wcscat]
        start = 1
    else:
        start = 0

    if not (hasattr(wcscat, '__iter__') and all(isinstance(wcat, TPWCS)
                                                for wcat in wcscat[start:])):
        raise TypeError("Input 'wcscat' must be either a single TPWCS-derived "
                        " object or a list of TPWCS-derived objects.")

    wcs_im_cats = []
    for wcat in wcscat:
        if wcat.meta.get('catalog', None) is None:
            raise ValueError("Each object in 'wcscat' must have a valid "
                             "catalog.")

        wcs_im_cat = WCSImageCatalog(
            catalog=wcat.meta['catalog'],
            tpwcs=wcat,
            name=wcat.meta.get('name', 'Unknown'),
            group_id=wcat.meta.get('group_id', None)
        )
        wcs_im_cat.fit_status = "FAILED: Unknown error"

        wcs_im_cats.append(wcs_im_cat)

    # check fitgeom:
    fitgeom = fitgeom.lower()
    try:
        if minobj is None:
            minobj = SUPPORTED_FITGEOM_MODES[fitgeom]
            log.debug(f"Setting 'minobj' to {minobj} for fitgeom='{fitgeom}'")
    except KeyError:
        raise ValueError(
            f"Unsupported 'fitgeom'. Valid values are: {_SUPPORTED_FITGEOM_EN_STR:s}"
        )

    # process reference catalog or image if provided:
    if refcat is not None:
        if isinstance(refcat, TPWCS):
            if 'catalog' not in refcat.meta:
                raise ValueError("Reference 'TPWCS' must contain a "
                                 "catalog.")

            rcat = refcat.meta['catalog'].copy()

            if not ('RA' in rcat.colnames and 'DEC' in rcat.colnames):  # pragma: no branch
                # convert image x & y to world coordinates:
                ra, dec = refcat.det_to_world(rcat['x'], rcat['y'])
                rcat['RA'] = ra
                rcat['DEC'] = dec

            refcat_name = refcat.meta.get(
                'name', rcat.meta.get('name', 'Unnamed')
            )

            refcat = RefCatalog(rcat, name=refcat_name)

        elif isinstance(refcat, astropy.table.Table):
            if 'RA' not in refcat.colnames or 'DEC' not in refcat.colnames:
                raise KeyError("Reference catalogs *must* contain *both* 'RA' "
                               "and 'DEC' columns.")
            refcat = RefCatalog(
                refcat, name=refcat.meta.get('name', 'Unnamed')
            )

        else:
            raise TypeError("Unsupported 'refcat' type. Supported 'refcat' "
                            "types are 'tweakwcs.tpwcs.TPWCS' and "
                            "'astropy.table.Table'")

    # find group ID and assign images to groups:
    grouped_images = collections.defaultdict(list)
    for wcat in wcs_im_cats:
        grouped_images[wcat.group_id].append(wcat)

    # create WCSImageCatalog and WCSGroupCatalog:
    wcs_gcat = []
    for group_id, wcatalogs in grouped_images.items():
        if group_id is None:
            for wcat in wcatalogs:
                wcs_gcat.append(
                    WCSGroupCatalog(wcat, name='GROUP ID: None')
                )

        else:
            gcat = WCSGroupCatalog(
                wcatalogs,
                name='GROUP ID: {}'.format(group_id)
            )
            if not len(gcat.catalog):
                log.warning("Group with ID '{}' will not be aligned: empty "
                            "source catalog".format(group_id))

                for wcat in wcatalogs:
                    wcat.tpwcs.meta['fit_info'] = {
                        'status': 'FAILED: empty source catalog'
                    }

                continue

            wcs_gcat.append(gcat)

    # check that we have enough input images:
    if (refcat is None and len(wcs_gcat) < 2) or len(wcs_gcat) == 0:
        raise ValueError("Too few input images (or groups of images) with "
                         "non-empty catalogs.")

    # get the first image to be aligned and
    # create reference catalog if needed:
    if refcat is None:
        # create reference catalog:
        ref_imcat, current_wcat = max_overlap_pair(
            images=wcs_gcat,
            enforce_user_order=enforce_user_order or not expand_refcat
        )
        log.info("Selected image '{}' as reference image"
                 .format(ref_imcat.name))

        refcat = RefCatalog(ref_imcat.catalog, name=ref_imcat[0].name)

        for wcat in ref_imcat:
            wcat.tpwcs.meta['fit_info'] = {'status': 'REFERENCE'}

    else:
        # find the first image to be aligned:
        current_wcat = max_overlap_image(
            refimage=refcat,
            images=wcs_gcat,
            enforce_user_order=enforce_user_order or not expand_refcat
        )

    while current_wcat is not None:
        log.info("Aligning image catalog '{}' to the reference catalog."
                 .format(current_wcat.name))

        current_wcat.align_to_ref(
            refcat=refcat,
            ref_tpwcs=ref_tpwcs,
            match=match,
            minobj=minobj,
            fitgeom=fitgeom,
            nclip=nclip,
            sigma=sigma
        )

        for wcat in current_wcat:
            wcat.tpwcs.meta['fit_info'] = wcat.fit_info

        # add unmatched sources to the reference catalog:
        if expand_refcat:
            unmatched_src = current_wcat.get_unmatched_cat()
            refcat.expand_catalog(unmatched_src)
            log.info("Added {:d} unmatched sources from '{}' to the reference "
                     "catalog.".format(len(unmatched_src), current_wcat.name))

        # find the next image to be aligned:
        current_wcat = max_overlap_image(
            refimage=refcat,
            images=wcs_gcat,
            enforce_user_order=enforce_user_order or not expand_refcat
        )

    # log running time:
    runtime_end = datetime.now()
    log.info(" ")
    log.info("***** {:s}.{:s}() ended on {}"
             .format(__name__, function_name, runtime_end))
    log.info("***** {:s}.{:s}() TOTAL RUN TIME: {}"
             .format(__name__, function_name, runtime_end - runtime_begin))
    log.info(" ")

    eff_refcat = refcat.catalog
    return eff_refcat


def overlap_matrix(images):
    """
    Compute overlap matrix: non-diagonal elements (i,j) of this matrix are
    absolute value of the area of overlap on the sky between i-th input image
    and j-th input image.

    .. note::
        The diagonal of the returned overlap matrix is set to ``0.0``, i.e.,
        this function does not compute the area of the footprint of a single
        image on the sky.

    Parameters
    ----------

    images: list of WCSImageCatalog, WCSGroupCatalog, or RefCatalog
        A list of catalogs that implement :py:meth:`intersection_area` method.

    Returns
    -------
    m: numpy.ndarray
        A `numpy.ndarray` of shape ``NxN`` where ``N`` is equal to the
        number of input images. Each non-diagonal element (i,j) of this matrix
        is the absolute value of the area of overlap on the sky between i-th
        input image and j-th input image. Diagonal elements are set to ``0.0``.

    """
    nimg = len(images)
    m = np.zeros((nimg, nimg), dtype=np.double)
    for i in range(nimg):
        for j in range(i + 1, nimg):
            area = images[i]._guarded_intersection_area(images[j])
            m[j, i] = area
            m[i, j] = area
    return m


def max_overlap_pair(images, enforce_user_order):
    """
    Return a pair of images with the largest overlap.

    .. warning::
        Returned pair of images is "poped" from input ``images`` list and
        therefore on return ``images`` will contain a smaller number of
        elements.

    Parameters
    ----------

    images: list of WCSImageCatalog, WCSGroupCatalog, or RefCatalog
        A list of catalogs that implement :py:meth:`intersection_area` method.

    enforce_user_order: bool
        When ``enforce_user_order`` is `True`, a pair of images will be
        returned **in the same order** as they were arranged in the ``images``
        input list. That is, image overlaps will be ignored.

    Returns
    -------
    (im1, im2)
        Returns a tuple of two images - elements of input ``images`` list.
        When ``enforce_user_order`` is `True`, images are returned in the
        order in which they appear in the input ``images`` list. When the
        number of input images is smaller than two, ``im1`` and ``im2`` may
        be `None`.

    """
    nimg = len(images)

    if nimg == 0:
        return None, None

    elif nimg == 1:
        return images[0], None

    elif nimg == 2 or enforce_user_order:
        # for the special case when only two images are provided
        # return (refimage, image) in the same order as provided in 'images'.
        # Also, when ref. catalog is static - revert to old tweakreg behavior
        im1 = images.pop(0)  # reference image
        im2 = images.pop(0)
        return im1, im2

    m = overlap_matrix(images)
    i, j = np.unravel_index(m.argmax(), m.shape)
    si = np.sum(m[i])
    sj = np.sum(m[:, j])

    if si < sj:  # pragma: no branch
        i, j = j, i

    if i < j:  # pragma: no branch
        j -= 1

    im1 = images.pop(i)  # reference image
    im2 = images.pop(j)

    # Sort the remaining of the input list of images by overlap area
    # with the reference image (in decreasing order):
    row = m[i]
    row = np.delete(row, i)
    row = np.delete(row, j)
    sorting_indices = np.argsort(row)[::-1]
    sorted_images = [images[k] for k in sorting_indices]  # apply argsort
    del images[:]
    for im in sorted_images:
        images.append(im)

    return im1, im2


def max_overlap_image(refimage, images, enforce_user_order):
    """
    Return the image from the input ``images`` list that has the largest
    overlap with the ``refimage`` image.

    .. warning::
        Returned image of images is "poped" from input ``images`` list and
        therefore on return ``images`` will contain a smaller number of
        elements.

    Parameters
    ----------
    refimage: RefCatalog
        Reference catalog.

    images: list of WCSImageCatalog, or WCSGroupCatalog
        A list of catalogs that implement :py:meth:`intersection_area` method.

    enforce_user_order: bool
        When ``enforce_user_order`` is `True`, returned image is the first
        image from the ``images`` input list regardless ofimage overlaps.

    Returns
    -------
    image: WCSImageCatalog, WCSGroupCatalog, or None
        Returns an element of input ``images`` list. When input list is
        empty - `None` is returned.

    """
    if not images:
        return None

    if enforce_user_order:
        # revert to old tweakreg behavior
        return images.pop(0)

    idx = np.argmax([refimage.intersection_area(im) for im in images])

    return images.pop(idx)



"""
A module containing unit tests for the `linalg` module.

Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
import pytest
import numpy as np

from tweakwcs import linalg, linearfit


@pytest.mark.parametrize('tp,expected', [
    (np.float16, False),
    (np.longdouble, True),
])
def test_longdouble_cmp(tp, expected):
    assert linalg._is_longdouble_lte_flt_type(tp) == expected


def test_inv_order2():
    feps = 1000 * np.finfo(np.double).eps

    # Generate a 2D rotation/scale/skew matrix + some random noise.
    # The reason for this complicated approach is to avoid creating
    # singular matrices.
    angle1 = 360 * np.random.random()
    angle2 = angle1 + 50 * (np.random.random() - 0.5)
    scale1 = 0.7 + 0.6 * np.random.random()
    scale2 = 0.7 + 0.6 * np.random.random()
    a = linearfit.build_fit_matrix((angle1, angle2), (scale1, scale2))

    # invert using numpy.linalg:
    use_numpy = linalg._USE_NUMPY_LINALG_INV
    linalg._USE_NUMPY_LINALG_INV = True

    try:
        x = linalg.inv(a)
        r = np.identity(2) - np.dot(a, x)

        # Use Morris Newman's formula to asses the quality of the inversion
        # (see https://nvlpubs.nist.gov/nistpubs/jres/78B/jresv78Bn2p65_A1b.pdf
        #  January 3, 1974).
        err = 2.0 * np.abs(np.dot(x, r)).max() / (1.0 - np.abs(r).max())
        assert err < feps

    finally:
        linalg._USE_NUMPY_LINALG_INV = use_numpy

    # invert using tweakwcs.linalg:
    linalg._USE_NUMPY_LINALG_INV = False

    try:
        x = linalg.inv(a)
        r = np.identity(2) - np.dot(a, x)

        # Use Morris Newman's formula to asses the quality of the inversion
        # (see https://nvlpubs.nist.gov/nistpubs/jres/78B/jresv78Bn2p65_A1b.pdf
        #  January 3, 1974).
        err = 2.0 * np.abs(np.dot(x, r)).max() / (1.0 - np.abs(r).max())
        assert err < feps

    finally:
        linalg._USE_NUMPY_LINALG_INV = use_numpy


@pytest.mark.parametrize('use_numpy_inv', [True, False])
def test_inv_nonsquare(use_numpy_inv):
    use_numpy = linalg._USE_NUMPY_LINALG_INV
    linalg._USE_NUMPY_LINALG_INV = use_numpy_inv
    with pytest.raises(np.linalg.LinAlgError):
        try:
            linalg.inv(np.empty((1, 2)))
        finally:
            linalg._USE_NUMPY_LINALG_INV = use_numpy


@pytest.mark.parametrize('use_numpy_inv', [True, False])
def test_inv_singular(use_numpy_inv):
    use_numpy = linalg._USE_NUMPY_LINALG_INV
    linalg._USE_NUMPY_LINALG_INV = use_numpy_inv
    arr = np.array([[1.0, 1.0], [2.0, 2.0]])
    with pytest.raises(np.linalg.LinAlgError):
        try:
            linalg.inv(arr)
        finally:
            linalg._USE_NUMPY_LINALG_INV = use_numpy


@pytest.mark.parametrize('use_numpy_inv', [True, False])
def test_inv_nan(use_numpy_inv):
    use_numpy = linalg._USE_NUMPY_LINALG_INV
    linalg._USE_NUMPY_LINALG_INV = use_numpy_inv
    arr = np.array([[1.0, 1.0, -1.], [2.0, -2.1, 4.0], [-1.0, np.nan, 1.0]])
    with pytest.raises(np.linalg.LinAlgError):
        try:
            linalg.inv(arr)
        finally:
            linalg._USE_NUMPY_LINALG_INV = use_numpy


def test_inv_high_dim():
    use_numpy = linalg._USE_NUMPY_LINALG_INV
    linalg._USE_NUMPY_LINALG_INV = False
    arr = np.random.random((4, 4, 4))
    with pytest.raises(np.linalg.LinAlgError):
        try:
            linalg.inv(arr)
        finally:
            linalg._USE_NUMPY_LINALG_INV = use_numpy



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
A module that provides utility functions for WCS transformations.

:Authors: Mihai Cara (contact: help@stsci.edu)

"""
# NOTES:
#
# Currently this module implements some general algorithms from
# jwst.transforms.tpcorr module so that tweakwcs would not have a
# hard dependency on jwst pipeline.

import math
import numpy as np

from . import __version__  # noqa: F401


__all__ = ['planar_rot_3d']

__author__ = 'Mihai Cara'


def planar_rot_3d(angle, axis):
    """
    Create a 3D rotation matrix that performs a rotation *in a plane*
    perpendicular to the specified ``axis``.

    """
    if axis not in range(3):
        raise ValueError("'axis' must be either 0, 1, or 2.")
    axis = int(axis)
    cs = math.cos(angle)
    sn = math.sin(angle)
    axisv = np.array(axis * [0.0] + [1.0] + (2 - axis) * [0.0],
                     dtype=np.double)
    mat_2d = np.array([[cs, sn], [-sn, cs]], dtype=np.double)
    return np.insert(np.insert(mat_2d, axis, [0.0, 0.0], 1), axis, axisv, 0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides support for manipulating tangent-plane corrections
of ``WCS``.

:Authors: Mihai Cara

:License: :doc:`../LICENSE`

"""
import logging
from copy import deepcopy
from abc import ABC, abstractmethod
from distutils.version import LooseVersion

import numpy as np
import astropy

try:
    import gwcs
    if LooseVersion(gwcs.__version__) > '0.12.0':
        from gwcs.geometry import CartesianToSpherical, SphericalToCartesian
        _GWCS_VER_GT_0P12 = True
    else:
        _GWCS_VER_GT_0P12 = False
except ImportError:
    _GWCS_VER_GT_0P12 = False

if LooseVersion(astropy.__version__) >= '4.0':
    _ASTROPY_VER_GE_4 = True
    from astropy.modeling import CompoundModel
    from astropy.modeling.models import (
        AffineTransformation2D, Scale, Identity, Mapping, Const1D,
        RotationSequence3D
    )
else:
    _ASTROPY_VER_GE_4 = False

from .linalg import inv
from . import __version__  # noqa: F401

__author__ = 'Mihai Cara'

__all__ = ['TPWCS', 'JWSTgWCS', 'FITSWCS']

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

_RAD2ARCSEC = 3600.0 * np.rad2deg(1.0)
_ARCSEC2RAD = 1.0 / _RAD2ARCSEC


def _tp2tp(tpwcs1, tpwcs2, s=None):
    x = np.array([-0.5, 0.5, -0.5, 0.0], dtype=np.double)
    y = np.array([-0.5, -0.5, 0.5, 0.0], dtype=np.double)

    if s is None:
        if 'fit_info' in tpwcs1.meta and 'center' in tpwcs1.meta['fit_info']:
            cx, cy = tpwcs1.meta['fit_info']['center']
            s = tpwcs1.tanp_pixel_scale(cx, cy)
        else:
            if tpwcs2.wcs.pixel_bounds is None:
                # TODO: A possible improvement would be to get an estimate
                #       of "center" (where scale is estimated) from source
                #       positions (if any).
                xt, yt = tpwcs1.world_to_tanp(*tpwcs2.det_to_world(x, y))
            else:
                cx, cy = np.mean(tpwcs2.wcs.pixel_bounds, axis=1)
                xt, yt = tpwcs1.world_to_tanp(
                    *tpwcs2.det_to_world(cx + x, cy + y)
                )
            m = np.array([(xt[1:-1] - xt[0]), (yt[1:-1] - yt[0])])
            s = np.sqrt(np.fabs(np.linalg.det(m)))

    xrp, yrp = tpwcs2.world_to_tanp(*tpwcs1.tanp_to_world(x * s, y * s))

    xrp = np.array(xrp, np.longdouble)
    yrp = np.array(yrp, np.longdouble)

    matrix = np.array([(xrp[1:-1] - xrp[0]), (yrp[1:-1] - yrp[0])]) / s
    shift = np.array([xrp[-1], yrp[-1]])

    return matrix, shift


class TPWCS(ABC):
    """ A class that provides common interface for manipulating WCS information
    and for managing tangent-plane corrections.

    """

    def __init__(self, wcs, meta=None):
        """
        Parameters
        ----------
        wcs: ``WCS object``
            A `WCS` object supported by a child class.

        meta: dict, None, optional
            Dictionary that will be merged to the object's ``meta`` fields.

        """
        self._owcs = wcs
        self._wcs = deepcopy(wcs)
        self._meta = {} if meta is None else dict(meta)

    @property
    def wcs(self):
        """ Get current WCS object. """
        return self._wcs

    @property
    def original_wcs(self):
        """ Get original WCS object. """
        return self._owcs

    def copy(self):
        """ Returns a deep copy of this object. """
        return deepcopy(self)

    @abstractmethod
    def set_correction(self, matrix=[[1, 0], [0, 1]], shift=[0, 0],
                       ref_tpwcs=None, meta=None, **kwargs):
        """
        Sets a tangent-plane correction of the WCS object according to
        the provided liniar parameters. In addition, this function updates
        the ``meta`` attribute of the `TPWCS`-derived object with the values
        of keyword arguments except for the argument ``meta`` which is
        *merged* with the *attribute* ``meta``.

        Parameters
        ----------
        matrix: list, numpy.ndarray, optional
            A ``2x2`` array or list of lists coefficients representing scale,
            rotation, and/or skew transformations.

        shift: list, numpy.ndarray, optional
            A list of two coordinate shifts to be applied to coordinates
            *after* ``matrix`` transformations are applied.

        ref_tpwcs: TPWCS, None, optional
            A reference WCS of the type ``TPWCS`` that provides the tangent
            plane in which corrections (``matrix`` and ``shift``) were defined.
            When not provided (i.e., set to `None`), it is assumed that the
            transformations are being applied directly to *this* image WCS'
            tangent plane.

        meta: dict, None, optional
            Dictionary that will be merged to the object's ``meta`` fields.

        **kwargs: optional parameters
            Optional parameters for the WCS corrector.

        """
        # save linear transformation info to the meta attribute:
        self._meta['matrix'] = np.array(matrix, dtype=np.double)
        self._meta['shift'] = shift
        if meta is not None:
            self._meta.update(meta)

    @abstractmethod
    def det_to_world(self, x, y):
        """
        Convert pixel coordinates to sky coordinates using full
        (i.e., including distortions) transformations.

        """
        ra, dec = x, y
        return ra, dec

    @abstractmethod
    def world_to_det(self, ra, dec):
        """
        Convert sky coordinates to image's pixel coordinates using full
        (i.e., including distortions) transformations.

        """
        x, y = ra, dec
        return x, y

    @abstractmethod
    def det_to_tanp(self, x, y):
        """
        Convert detector (pixel) coordinates to tangent plane coordinates.

        """
        return x, y

    @abstractmethod
    def tanp_to_det(self, x, y):
        """
        Convert tangent plane coordinates to detector (pixel) coordinates.

        """
        return x, y

    @abstractmethod
    def world_to_tanp(self, ra, dec):
        """
        Convert tangent plane coordinates to detector (pixel) coordinates.

        """
        x, y = ra, dec
        return x, y

    @abstractmethod
    def tanp_to_world(self, x, y):
        """
        Convert tangent plane coordinates to world coordinates.

        """
        ra, dec = x, y
        return ra, dec

    def tanp_pixel_scale(self, x, y):
        """ Estimate pixel scale in the tangent plane near a location
        in the detector's coordinate system given by parameters ``x`` and
        ``y``.

        Parameters
        ----------
        x: int, float
            X-coordinate of the pixel in the detector's coordinate system
            near which pixel scale in the tangent plane needs to be estimated.

        y: int, float
            Y-coordinate of the pixel in the detector's coordinate system
            near which pixel scale in the tangent plane needs to be estimated.

        Returns
        -------

        pscale: float
            Pixel scale [in units used in the tangent plane coordinate system]
            in the tangent plane near a location specified by detector
            coordinates ``x`` and ``y``.

        """
        xs = np.asarray([x - 0.5, x - 0.5, x + 0.5, x + 0.5])
        ys = np.asarray([y - 0.5, y + 0.5, y + 0.5, y - 0.5])
        xt, yt = self.det_to_tanp(xs, ys)
        area = 0.5 * np.abs(
            xt[0] * yt[1] + xt[1] * yt[2] + xt[2] * yt[3] + xt[3] * yt[0] -
            xt[1] * yt[0] - xt[2] * yt[1] - xt[3] * yt[2] - xt[0] * yt[3]
        )
        pscale = float(np.sqrt(area))
        return pscale

    @property
    def tanp_center_pixel_scale(self):
        """ Estimate pixel scale in the tangent plane near a location
        in the detector's coordinate system corresponding to the origin of the
        tangent plane.

        Returns
        -------
        pscale: float
            Pixel scale [in units used in the tangent plane coordinate system]
            in the tangent plane near a location in the detector's plane
            corresponding to the origin of the tangent plane.

        """
        pscale = self._get_tanp_center_pixel_scale()
        return pscale

    def _get_tanp_center_pixel_scale(self):
        x, y = self.tanp_to_det(0.0, 0.0)
        pscale = self.tanp_pixel_scale(x, y)
        return pscale

    @property
    def meta(self):
        return self._meta

    @property
    def bounding_box(self):
        """
        Get the bounding box (if any) of the underlying image for which
        the original WCS is defined.

        """
        return None


class FITSWCS(TPWCS):
    """ A class for holding ``FITS`` ``WCS`` information and for managing
    tangent-plane corrections.

    .. note::
        Currently only WCS objects that have ``CPDIS``, ``DET2IM``, and ``SIP``
        distortions *before* the application of the ``CD`` or ``PC`` matrix are
        supported.

    """
    def __init__(self, wcs, meta=None):
        """
        Parameters
        ----------

        wcs: astropy.wcs.WCS
            An `astropy.wcs.WCS` object.

        """
        valid, message = self._check_wcs_structure(wcs)
        if not valid:
            raise ValueError("Unsupported WCS structure: " + message)

        super().__init__(wcs=wcs, meta=meta)
        wcslin = wcs.deepcopy()

        # strip all *known* distortions:
        wcslin.cpdis1 = None
        wcslin.cpdis2 = None
        wcslin.det2im1 = None
        wcslin.det2im2 = None
        wcslin.sip = None
        wcslin.wcs.set()

        self._wcslin = wcslin

    def _check_wcs_structure(self, wcs):
        """
        Attempt detecting unknown distortion corrections. We basically
        want to make sure that we can turn off all distortions that are
        happening between detector's plane and the intermediate coordinate
        plane. This is necessary until we can find a better way of getting
        from intermediate coordinates to world coordinates.

        """
        if wcs is None:
            return False, "WCS cannot be None."

        if not wcs.is_celestial:
            return False, "WCS must be exclusively a celestial WCS."

        wcs = wcs.deepcopy()
        naxis1, naxis2 = wcs.pixel_shape

        # check mapping of corners and CRPIX:
        pts = np.array([[1.0, 1.0], [1.0, naxis2], [naxis1, 1.0],
                        [naxis1, naxis2], wcs.wcs.crpix])

        sky_all = wcs.all_pix2world(pts, 1)
        foc_all = wcs.pix2foc(pts, 1)

        # strip all *known* distortions:
        wcs.cpdis1 = None
        wcs.cpdis2 = None
        wcs.det2im1 = None
        wcs.det2im2 = None
        wcs.sip = None

        # check that pix2foc includes no other distortions besides the ones
        # that we have turned off above:
        if not np.allclose(pts, wcs.pix2foc(pts, 1)):
            False, "'pix2foc' contains unknown distortions"

        wcs.wcs.set()

        # check that pix2foc contains all known distortions:
        if not np.allclose(wcs.all_world2pix(sky_all, 1), foc_all, atol=1e-3,
                           rtol=0):
            return False, "'WCS.pix2foc()' does not include all distortions."

        return True, ''

    def _get_tanp_center_pixel_scale(self):
        pscale = self.tanp_pixel_scale(*self._wcs.wcs.crpix)
        return pscale

    def set_correction(self, matrix=[[1, 0], [0, 1]], shift=[0, 0],
                       ref_tpwcs=None, meta=None, **kwargs):
        """
        Computes a corrected (aligned) wcs based on the provided linear
        transformation. In addition, this function updates the ``meta``
        attribute of the `FITSWCS` object with the the values of keyword
        arguments except for the argument ``meta`` which is *merged* with
        the *attribute* ``meta``.

        Parameters
        ----------
        matrix: list, numpy.ndarray
            A ``2x2`` array or list of lists coefficients representing scale,
            rotation, and/or skew transformations.

        shift: list, numpy.ndarray
            A list of two coordinate shifts to be applied to coordinates
            *after* ``matrix`` transformations are applied.

        ref_tpwcs: TPWCS, None, optional
            A reference WCS of the type ``TPWCS`` that provides the tangent
            plane in which corrections (``matrix`` and ``shift``) were defined.
            When not provided (i.e., set to `None`), it is assumed that the
            transformations are being applied directly to *this* image WCS'
            tangent plane.

        meta: dict, None, optional
            Dictionary that will be merged to the object's ``meta`` fields.

        **kwargs: optional parameters
            Optional parameters for the WCS corrector. `FITSWCS` ignores these
            arguments (except for storing them in the ``meta`` attribute).

        """
        wcs = self._wcs
        orig_wcs = wcs.deepcopy()
        if ref_tpwcs is None:
            ref_tpwcs = FITSWCS(wcs.deepcopy())

        naxis1, naxis2 = wcs.pixel_shape

        shift = -np.dot(inv(matrix), shift)

        # estimate step for numerical differentiation. We need a step
        # large enough to avoid rounding errors and small enough to get a
        # better precision for numerical differentiation.
        # TODO: The logic below should be revised at a later time so that it
        # better takes into account the two competing requirements.
        hx = max(1.0, min(10, (wcs.wcs.crpix[0] - 1.0) / 100.0,
                          (naxis1 - wcs.wcs.crpix[0]) / 100.0))
        hy = max(1.0, min(10, (wcs.wcs.crpix[1] - 1.0) / 100.0,
                          (naxis2 - wcs.wcs.crpix[1]) / 100.0))

        # compute new CRVAL for the image WCS:
        crpix2d = np.atleast_2d(wcs.wcs.crpix)
        crval = wcs.wcs_pix2world(crpix2d, 1).ravel()
        crpixinref = ref_tpwcs.world_to_tanp(*crval)

        crpixinref = np.dot(
            np.subtract(crpixinref, shift),
            np.transpose(matrix)
        ).astype(np.double)
        wcs.wcs.crval = ref_tpwcs.tanp_to_world(crpixinref[0], crpixinref[1])
        wcs.wcs.set()

        # initial approximation for CD matrix of the image WCS:
        (U, u) = self._linearize(orig_wcs, ref_tpwcs, wcs.wcs.crpix,
                                 matrix, shift, hx=hx, hy=hy)
        if hasattr(wcs.wcs, 'pc'):
            wcs.wcs.pc = np.dot(wcs.wcs.pc, U).astype(np.double)
        else:
            wcs.wcs.cd = np.dot(wcs.wcs.cd, U).astype(np.double)
        wcs.wcs.set()

        # save linear transformation info to the meta attribute:
        super().set_correction(matrix=matrix, shift=shift, meta=meta, **kwargs)

    def det_to_world(self, x, y):
        """
        Convert pixel coordinates to sky coordinates using full
        (i.e., including distortions) transformations.

        """
        ra, dec = self._wcs.all_pix2world(x, y, 0)
        return ra, dec

    def world_to_det(self, ra, dec):
        """
        Convert sky coordinates to image's pixel coordinates using full
        (i.e., including distortions) transformations.

        """
        x, y = self._wcs.all_world2pix(ra, dec, 0, tolerance=1e-6, maxiter=50)
        return x, y

    def det_to_tanp(self, x, y):
        """
        Convert detector (pixel) coordinates to tangent plane coordinates.

        """
        x, y = self._wcs.pix2foc(x, y, 0)
        return x, y

    def tanp_to_det(self, x, y):
        """
        Convert tangent plane coordinates to detector (pixel) coordinates.

        """
        ndim = np.ndim(x)
        ra, dec = self._wcs.wcs_pix2world(x, y, 0)
        if not ndim:
            # deal with a bug/feature in stwcs' all_world2pix v.1.5.3
            ra = np.atleast_1d(ra)
            dec = np.atleast_1d(dec)
        x, y = self._wcs.all_world2pix(ra, dec, 0, tolerance=1e-6, maxiter=50)
        if not ndim:
            x = float(x)
            y = float(y)
        return x, y

    def world_to_tanp(self, ra, dec):
        """
        Convert tangent plane coordinates to detector (pixel) coordinates.

        """
        x, y = self._wcs.wcs_world2pix(ra, dec, 0)
        return x, y

    def tanp_to_world(self, x, y):
        """ Convert tangent plane coordinates to world coordinates. """
        ra, dec = self._wcs.wcs_pix2world(x, y, 0)
        return ra, dec

    @property
    def bounding_box(self):
        """
        Get the bounding box (if any) of the underlying image for which
        the original WCS is defined.

        """
        if self._owcs.pixel_bounds is None:
            if self._owcs.pixel_shape is not None:
                nx, ny = self._owcs.pixel_shape
            elif self._owcs.array_shape is not None:
                ny, nx = self._owcs.array_shape
            else:
                return None

            return ((-0.5, nx - 0.5), (-0.5, ny - 0.5))

        else:
            return self._owcs.pixel_bounds

    def _linearize(self, wcsima, ref_tpwcs, imcrpix, f, shift, hx=1.0, hy=1.0):
        """ linearization using 5-point formula for first order derivative. """
        x0 = imcrpix[0] - 1
        y0 = imcrpix[1] - 1

        px = np.asarray([x0, x0 - hx, x0 - hx * 0.5, x0 + hx * 0.5,
                         x0 + hx, x0, x0, x0, x0], dtype=np.double)
        py = np.asarray([y0, y0, y0, y0, y0, y0 - hy, y0 - hy * 0.5,
                         y0 + hy * 0.5, y0 + hy], dtype=np.double)

        # convert image coordinates to reference image coordinates:
        ra, dec = wcsima.wcs_pix2world(px, py, 0)
        px, py = np.array(ref_tpwcs.world_to_tanp(ra, dec))

        # apply linear fit transformation:
        px, py = np.dot(f, [px - shift[0], py - shift[1]]).astype(np.double)
        # convert back to image coordinate system:
        p = np.array(self.wcs.wcs_world2pix(*ref_tpwcs.tanp_to_world(px, py), 0),
                     dtype=np.longdouble).T

        # derivative with regard to x:
        u1 = ((p[1] - p[4]) + 8 * (p[3] - p[2])) / (6 * hx)
        # derivative with regard to y:
        u2 = ((p[5] - p[8]) + 8 * (p[7] - p[6])) / (6 * hy)

        return (np.asarray([u1, u2]).T, p[0])


def _get_submodel(model, name):
    """ Return the first occurence of a sub-model. Search is performed by
        model name.
    """
    if not isinstance(model, CompoundModel):
        return model if model.name == name else None

    for m in model.traverse_postorder():
        if m.name == name:
            return m

    return None


class JWSTgWCS(TPWCS):
    """ A class for holding ``JWST`` ``gWCS`` information and for managing
    tangent-plane corrections.

    """
    def __init__(self, wcs, wcsinfo, meta=None):
        """
        Parameters
        ----------
        wcs: GWCS
            A `GWCS` object.

        wcsinfo: dict
            A dictionary containing reference angles of ``JWST`` instrument
            provided in the ``wcsinfo`` property of ``JWST`` meta data.

            This dictionary **must contain** the following keys and values:

            'v2_ref': float
                V2 position of the reference point in arc seconds.

            'v3_ref': float
                V3 position of the reference point in arc seconds.

            'roll_ref': float
                Roll angle in degrees.

        meta: dict, None, optional
            Dictionary that will be merged to the object's ``meta`` fields.

        """
        if not _ASTROPY_VER_GE_4:
            raise NotImplementedError(
                "JWST support requires astropy version >= 4.0"
            )

        if not _GWCS_VER_GT_0P12:
            raise NotImplementedError(
                "JWST support requires gwcs version > 0.12.0 "
                "To pip install minimal required version, do the following:\n"
                "pip install git+https://github.com/spacetelescope/gwcs@"
                "1c1cb3bb35caddef80fb760ea68bc71e189d32de"
            )

        valid, message = self._check_wcs_structure(wcs)
        if not valid:
            raise ValueError("Unsupported WCS structure: {}".format(message))

        super().__init__(wcs=wcs, meta=meta)

        v2_ref = wcsinfo['v2_ref']
        v3_ref = wcsinfo['v3_ref']
        roll_ref = wcsinfo['roll_ref']

        self._wcsinfo = {'v2_ref': v2_ref, 'v3_ref': v3_ref,
                         'roll_ref': roll_ref}

        # perform additional check that if tangent plane correction is already
        # present in the WCS pipeline, it is of TPCorr class and that
        # its parameters are consistent with reference angles:
        frms = wcs.available_frames
        if 'v2v3corr' in frms:
            self._v23name = 'v2v3corr'
            self._tpcorr = deepcopy(
                wcs.pipeline[frms.index('v2v3corr') - 1].transform
            )
            self._default_tpcorr = None
            if not JWSTgWCS._check_tpcorr_structure(self._tpcorr):
                raise ValueError("Unsupported tangent-plance correction type.")

            # check that transformation parameters are consistent with
            # reference angles:
            v2ref, v3ref, roll = self._tpcorr['det_to_optic_axis'].angles.value

            eps_v2 = 10.0 * np.finfo(v2_ref).eps
            eps_v3 = 10.0 * np.finfo(v3_ref).eps
            eps_roll = 10.0 * np.finfo(roll_ref).eps
            if not (np.isclose(v2_ref, v2ref * 3600.0, rtol=eps_v2) and
                    np.isclose(v3_ref, -v3ref * 3600.0, rtol=eps_v3) and
                    np.isclose(roll_ref, roll, rtol=eps_roll)):
                raise ValueError(
                    "WCS/TPCorr parameters 'v2ref', 'v3ref', and/or 'roll' "
                    "differ from the corresponding reference values."
                )
            self._partial_tpcorr = JWSTgWCS._v2v3_to_tpcorr_from_full(
                self._tpcorr
            )

        else:
            self._v23name = 'v2v3vacorr' if 'v2v3vacorr' in frms else 'v2v3'
            self._tpcorr = None
            self._default_tpcorr = JWSTgWCS._tpcorr_init(
                v2_ref=v2_ref / 3600.0,
                v3_ref=v3_ref / 3600.0,
                roll_ref=roll_ref
            )
            self._partial_tpcorr = JWSTgWCS._v2v3_to_tpcorr_from_full(
                self._default_tpcorr
            )

        self._update_transformations()

    @staticmethod
    def _check_tpcorr_structure(tpcorr):
        # implement a more sophisticated check later
        if tpcorr.name != 'JWST tangent-plane linear correction. v1':
            return False
        return True

    def _update_transformations(self):
        # define transformations from detector/world coordinates to
        # the tangent plane:
        detname = self._wcs.pipeline[0].frame.name
        worldname = self._wcs.pipeline[-1].frame.name

        # Generally needed transformations:
        self._world_to_v23 = self._wcs.get_transform(worldname, self._v23name)
        self._v23_to_world = self._wcs.get_transform(self._v23name, worldname)
        self._det_to_v23 = self._wcs.get_transform(detname, self._v23name)
        self._det_to_world = self._wcs.__call__

        # Optional / convenience transformations:
        try:
            self._v23_to_det = self._wcs.get_transform(self._v23name, detname)
        except NotImplementedError:
            self._v23_to_det = None

        try:
            self._world_to_det = self._wcs.invert
        except NotImplementedError:
            self._world_to_det = None

    @staticmethod
    def _tpcorr_combine_affines(tpcorr, matrix, shift):
        AffineTransformation2D(matrix, shift)  # check input parameters are OK
        m = np.dot(matrix, tpcorr['tp_affine'].matrix.value)
        t = np.dot(matrix, tpcorr['tp_affine'].translation.value) + shift
        tpcorr['tp_affine'].matrix = m
        tpcorr['tp_affine'].translation = t

        # update the affine transformation of the inverse model as well:
        invm = np.linalg.inv(m)
        tpcorr.inverse['tp_affine_inv'].matrix = invm
        tpcorr.inverse['tp_affine_inv'].translation = -np.dot(invm, t)

    @staticmethod
    def _tpcorr_init(v2_ref, v3_ref, roll_ref):
        s2c = SphericalToCartesian(name='s2c', wrap_lon_at=180)
        c2s = CartesianToSpherical(name='c2s', wrap_lon_at=180)

        unit_conv = Scale(1.0 / 3600.0, name='arcsec_to_deg_1D')
        unit_conv = unit_conv & unit_conv
        unit_conv.name = 'arcsec_to_deg_2D'

        unit_conv_inv = Scale(3600.0, name='deg_to_arcsec_1D')
        unit_conv_inv = unit_conv_inv & unit_conv_inv
        unit_conv_inv.name = 'deg_to_arcsec_2D'

        affine = AffineTransformation2D(name='tp_affine')
        affine_inv = AffineTransformation2D(name='tp_affine_inv')

        rot = RotationSequence3D(
            [v2_ref, -v3_ref, roll_ref],
            'zyx',
            name='det_to_optic_axis'
        )
        rot_inv = rot.inverse
        rot_inv.name = 'optic_axis_to_det'

        # projection submodels:
        c2tan = ((Mapping((0, 1, 2), name='xyz') /
                  Mapping((0, 0, 0), n_inputs=3, name='xxx')) |
                 Mapping((1, 2), name='xtyt'))
        c2tan.name = 'Cartesian 3D to TAN'
        tan2c = (Mapping((0, 0, 1), n_inputs=2, name='xtyt2xyz') |
                 (Const1D(1, name='one') & Identity(2, name='I(2D)')))
        tan2c.name = 'TAN to cartesian 3D'

        total_corr = (
            unit_conv | s2c | rot | c2tan | affine |
            tan2c | rot_inv | c2s | unit_conv_inv
        )
        total_corr.name = 'JWST tangent-plane linear correction. v1'

        inv_total_corr = (
            unit_conv | s2c | rot | c2tan | affine_inv |
            tan2c | rot_inv | c2s | unit_conv_inv
        )
        inv_total_corr.name = 'Inverse JWST tangent-plane linear correction. v1'

        # TODO
        # re-enable circular inverse definitions once
        # https://github.com/spacetelescope/asdf/issues/744 or
        # https://github.com/spacetelescope/asdf/issues/745 are resolved.
        #
        # inv_total_corr.inverse = total_corr
        total_corr.inverse = inv_total_corr

        return total_corr

    @staticmethod
    def _v2v3_to_tpcorr_from_full(tpcorr):
        s2c = tpcorr['s2c']
        c2s = tpcorr['c2s']

        # unit_conv = _get_submodel(tpcorr, 'arcsec_to_deg_2D')
        # unit_conv_inv = _get_submodel(tpcorr, 'deg_to_arcsec_2D')
        #
        # The code below is a work-around to the code above not working.
        # TODO: understand why _get_submodel is unable to retrieve
        #       some submodels in a compound model.
        #
        unit_conv = _get_submodel(tpcorr, 'arcsec_to_deg_1D')
        unit_conv = unit_conv & unit_conv
        unit_conv.name = 'arcsec_to_deg_2D'
        unit_conv_inv = _get_submodel(tpcorr, 'deg_to_arcsec_1D')
        unit_conv_inv = unit_conv_inv & unit_conv_inv
        unit_conv_inv.name = 'deg_to_arcsec_2D'

        affine = tpcorr['tp_affine']
        affine_inv = affine.inverse
        affine_inv.name = 'tp_affine_inv'

        rot = tpcorr['det_to_optic_axis']
        rot_inv = rot.inverse
        rot_inv.name = 'optic_axis_to_det'

        # c2tan = _get_submodel(tpcorr, 'Cartesian 3D to TAN')
        # tan2c = _get_submodel(tpcorr, 'TAN to cartesian 3D')
        #
        # The code below is a work-around to the code above not working.
        # TODO: understand why _get_submodel is unable to retrieve
        #       some submodels in a compound model.
        #
        c2tan = ((Mapping((0, 1, 2), name='xyz') /
                  Mapping((0, 0, 0), n_inputs=3, name='xxx')) |
                 Mapping((1, 2), name='xtyt'))
        c2tan.name = 'Cartesian 3D to TAN'
        tan2c = (Mapping((0, 0, 1), n_inputs=2, name='xtyt2xyz') |
                 (Const1D(1, name='one') & Identity(2, name='I(2D)')))
        tan2c.name = 'TAN to cartesian 3D'

        v2v3_to_tpcorr = unit_conv | s2c | rot | c2tan | affine
        v2v3_to_tpcorr.name = 'jwst_v2v3_to_tpcorr'

        tpcorr_to_v2v3 = affine_inv | tan2c | rot_inv | c2s | unit_conv_inv
        tpcorr_to_v2v3.name = 'jwst_tpcorr_to_v2v3'

        v2v3_to_tpcorr.inverse = tpcorr_to_v2v3
        tpcorr_to_v2v3.inverse = v2v3_to_tpcorr

        return v2v3_to_tpcorr

    @property
    def ref_angles(self):
        """ Return a ``wcsinfo``-like dictionary of main WCS parameters. """
        return {k: v for k, v in self._wcsinfo.items()}

    def set_correction(self, matrix=[[1, 0], [0, 1]], shift=[0, 0],
                       ref_tpwcs=None, meta=None, **kwargs):
        """
        Sets a tangent-plane correction of the GWCS object according to
        the provided liniar parameters. In addition, this function updates
        the ``meta`` attribute of the `JWSTgWCS` object with the values
        of keyword arguments except for the argument ``meta`` which is
        *merged* with the *attribute* ``meta``.

        Parameters
        ----------
        matrix: list, numpy.ndarray
            A ``2x2`` array or list of lists coefficients representing scale,
            rotation, and/or skew transformations.

        shift: list, numpy.ndarray
            A list of two coordinate shifts to be applied to coordinates
            *after* ``matrix`` transformations are applied.

        ref_tpwcs: TPWCS, None, optional
            A reference WCS of the type ``TPWCS`` that provides the tangent
            plane in which corrections (``matrix`` and ``shift``) were defined.
            When not provided (i.e., set to `None`), it is assumed that the
            transformations are being applied directly to *this* image WCS'
            tangent plane.

        meta: dict, None, optional
            Dictionary that will be merged to the object's ``meta`` fields.

        **kwargs: optional parameters
            Optional parameters for the WCS corrector. `JWSTgWCS` ignores these
            arguments (except for storing them in the ``meta`` attribute).

        """
        frms = self._wcs.available_frames

        if ref_tpwcs is None:
            matrix = np.array(matrix, dtype=np.double)
            shift = np.array(shift, dtype=np.double)
        else:
            # compute linear transformation from the tangent plane used for
            # alignment to the tangent plane of this wcs:
            r, t = _tp2tp(ref_tpwcs, self)
            matrix = np.linalg.multi_dot([r, matrix, inv(r)]).astype(np.double)
            shift = (np.dot(r, shift) - np.dot(matrix, t) + t).astype(np.double)

        # if original WCS did not have tangent-plane corrections, create
        # new correction and add it to the WCs pipeline:
        if self._tpcorr is None:
            self._tpcorr = JWSTgWCS._tpcorr_init(
                v2_ref=self._wcsinfo['v2_ref'] / 3600.0,
                v3_ref=self._wcsinfo['v3_ref'] / 3600.0,
                roll_ref=self._wcsinfo['roll_ref']
            )

            JWSTgWCS._tpcorr_combine_affines(
                self._tpcorr,
                matrix,
                _ARCSEC2RAD * np.asarray(shift)
            )

            self._partial_tpcorr = JWSTgWCS._v2v3_to_tpcorr_from_full(self._tpcorr)

            idx_v2v3 = frms.index(self._v23name)
            pipeline = deepcopy(self._wcs.pipeline)
            pf, pt = pipeline[idx_v2v3]
            pipeline[idx_v2v3] = (pf, deepcopy(self._tpcorr))
            frm_v2v3corr = deepcopy(pf)
            frm_v2v3corr.name = 'v2v3corr'
            pipeline.insert(idx_v2v3 + 1, (frm_v2v3corr, pt))
            self._wcs = gwcs.WCS(pipeline, name=self._owcs.name)
            self._v23name = 'v2v3corr'

        else:
            # combine old and new corrections into a single one and replace
            # old transformation with the combined correction transformation:
            JWSTgWCS._tpcorr_combine_affines(
                self._tpcorr,
                matrix,
                _ARCSEC2RAD * np.asarray(shift)
            )

            self._partial_tpcorr = JWSTgWCS._v2v3_to_tpcorr_from_full(self._tpcorr)

            idx_v2v3 = frms.index(self._v23name)
            pipeline = deepcopy(self._wcs.pipeline)
            pipeline[idx_v2v3 - 1] = (pipeline[idx_v2v3 - 1].frame,
                                      deepcopy(self._tpcorr))
            self._wcs = gwcs.WCS(pipeline, name=self._owcs.name)

        # reset definitions of the transformations from detector/world
        # coordinates to the tangent plane:
        self._update_transformations()

        # save linear transformation info to the meta attribute:
        super().set_correction(matrix=matrix, shift=shift, meta=meta, **kwargs)

    def _check_wcs_structure(self, wcs):
        if wcs is None or wcs.pipeline is None:
            return False, "Either WCS or its pipeline is None."

        frms = wcs.available_frames
        nframes = len(frms)

        if nframes < 3:
            return False, "There are fewer than 3 frames in the WCS pipeline."

        if frms.count(frms[0]) > 1 or frms.count(frms[-1]) > 1:
            return (False, "First and last frames in the WCS pipeline must "
                    "be unique.")

        if frms.count('v2v3') != 1 or frms.count('v2v3vacorr') > 1:
            return (False, "Only one 'v2v3' or 'v2v3vacorr' frame is allowed "
                    "in the WCS pipeline.")

        idx_v2v3 = frms.index('v2v3')
        if idx_v2v3 == 0 or idx_v2v3 == (nframes - 1):
            return (False, "'v2v3' frame cannot be first or last in the WCS "
                    "pipeline.")

        try:
            idx_v2v3vacorr = frms.index('v2v3vacorr')
            if idx_v2v3vacorr < idx_v2v3 or idx_v2v3vacorr == (nframes - 1):
                return (False, "'v2v3vacorr' frame cannot precede the 'v2v3' "
                        "frame or be the last frame in the WCS pipeline.")
            idx_v2v3 = idx_v2v3vacorr
        except ValueError:
            pass

        nv2v3corr = frms.count('v2v3corr')
        if nv2v3corr == 0:
            return True, ''
        elif nv2v3corr > 1:
            return (False, "Only one 'v2v3corr' correction frame is allowed "
                    "in the WCS pipeline.")

        idx_v2v3corr = frms.index('v2v3corr')
        if idx_v2v3corr != (idx_v2v3 + 1) or idx_v2v3corr == (nframes - 1):
            return (False, "'v2v3corr' frame is not in the correct position "
                    "in the WCS pipeline.")

        return True, ''

    def det_to_world(self, x, y):
        """
        Convert pixel coordinates to sky coordinates using full
        (i.e., including distortions) transformations.

        """
        ra, dec = self._det_to_world(x, y)
        return ra, dec

    def world_to_det(self, ra, dec):
        """
        Convert sky coordinates to image's pixel coordinates using full
        (i.e., including distortions) transformations.

        """
        if self._world_to_det is None:
            raise NotImplementedError("World to detector transformation has "
                                      "not been implemented.")
        x, y = self._world_to_det(ra, dec)
        return x, y

    def det_to_tanp(self, x, y):
        """
        Convert detector (pixel) coordinates to tangent plane coordinates.

        """
        v2, v3 = self._det_to_v23(x, y)
        x, y = self._partial_tpcorr(v2, v3)
        return _RAD2ARCSEC * x, _RAD2ARCSEC * y

    def tanp_to_det(self, x, y):
        """
        Convert tangent plane coordinates to detector (pixel) coordinates.

        """
        if self._partial_tpcorr.inverse is None or self._v23_to_det is None:
            raise NotImplementedError(
                "Tangent plane to detector transformation has not been "
                "implemented."
            )

        v2, v3 = self._partial_tpcorr.inverse(
            _ARCSEC2RAD * np.asanyarray(x),
            _ARCSEC2RAD * np.asanyarray(y)
        )
        x, y = self._v23_to_det(v2, v3)
        return x, y

    def world_to_tanp(self, ra, dec):
        """
        Convert tangent plane coordinates to detector (pixel) coordinates.

        """
        v2, v3 = self._world_to_v23(ra, dec)
        x, y = self._partial_tpcorr(v2, v3)
        return _RAD2ARCSEC * x, _RAD2ARCSEC * y

    def tanp_to_world(self, x, y):
        """ Convert tangent plane coordinates to world coordinates. """
        v2, v3 = self._partial_tpcorr.inverse(
            _ARCSEC2RAD * np.asanyarray(x),
            _ARCSEC2RAD * np.asanyarray(y)
        )
        ra, dec = self._v23_to_world(v2, v3)
        return ra, dec

    @property
    def bounding_box(self):
        """
        Get the bounding box (if any) of the underlying image for which
        the original WCS is defined.

        """
        if self._owcs.pixel_bounds is None:
            if self._owcs.pixel_shape is not None:
                nx, ny = self._owcs.pixel_shape
            elif self._owcs.array_shape is not None:
                ny, nx = self._owcs.array_shape
            else:
                return None

            return ((-0.5, nx - 0.5), (-0.5, ny - 0.5))

        else:
            return self._owcs.pixel_bounds



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides general purpose and/or specialized linear algebra
routines.

:Authors: Mihai Cara (contact: help@stsci.edu)

:License: :doc:`../LICENSE`

"""
# STDLIB
import logging

# THIRD-PARTY
import numpy as np

# LOCAL
from . import __version__  # noqa: F401

__author__ = 'Mihai Cara'

__all__ = ['inv']

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def _is_longdouble_lte_flt_type(flt_type=np.double):
    if np.longdouble is flt_type:
        return True

    fi1 = np.finfo(np.longdouble)
    fi2 = np.finfo(flt_type)

    lte_flt = all(
        [
            fi1.eps >= fi2.eps,
            fi1.bits <= fi2.bits,
            fi1.epsneg >= fi2.epsneg,
            fi1.iexp <= fi2.iexp,
            fi1.max <= fi2.max,
            fi1.maxexp <= fi2.maxexp,
            fi1.min >= fi2.min,
            fi1.minexp >= fi2.minexp,
            fi1.negep >= fi2.negep,
            fi1.nexp <= fi1.nexp,
            fi1.nmant <= fi2.nmant,
            fi1.precision <= fi2.precision,
            fi1.resolution >= fi2.resolution,
            fi1.tiny >= fi2.tiny,
        ]
    )

    return lte_flt


def _find_max_linalg_type():
    max_type = None
    for np_type in np.sctypes['float'][::-1]:  # pragma: no branch
        try:
            r = np.linalg.inv(np.identity(2, dtype=np_type))
            max_type = np_type
            eps = 100 * np.finfo(max_type).eps
            if np.max(np.abs(r - np.identity(2, dtype=np_type))) > eps:  # pragma: no branch
                log.warning('Loss of accuracy during matrix inversion.')
            break
        except TypeError:
            continue


_USE_NUMPY_LINALG_INV = _is_longdouble_lte_flt_type(flt_type=np.double)
_MAX_LINALG_TYPE = _find_max_linalg_type()
if _MAX_LINALG_TYPE is None:
    _USE_NUMPY_LINALG_INV = False  # pragma: no cover


def inv(m):
    """ This function computes inverse matrix using Gauss-Jordan elimination
    with full pivoting. Computations are performed using ``numpy.longdouble``
    precision. On systems on which ``numpy.longdouble`` is equivalent to
    ``numpy.double`` this function reverts to `numpy.linalg.inv` for
    performance reasons.

    Parameters
    ----------
    m: numpy.ndarray
        A 2D *square* matrix of type `numpy.ndarray`.

    Returns
    -------
    invm: numpy.ndarray
        Inverse matrix of the input matrix ``m``: a 2D *square* `numpy.ndarray`
        of type ``numpy.longdouble`` on systems on which it is more accurate
        than ``numpy.double``.

    """
    # check that matrix is square:
    if _USE_NUMPY_LINALG_INV:
        invm = np.linalg.inv(np.array(m, dtype=_MAX_LINALG_TYPE))
        # detect singularity:
        if not np.all(np.isfinite(invm)):
            raise np.linalg.LinAlgError('Singular matrix.')
        return invm

    m = np.array(m, dtype=np.longdouble)
    if len(m.shape) != 2 or m.shape[0] != m.shape[1]:
        raise np.linalg.LinAlgError("Input matrix must be a square matrix.")
    order = m.shape[0]

    # create permutation matrices:
    qt = np.eye(order, dtype=int)

    eps = np.finfo(np.double).tiny

    # initial inverse matrix:
    invm = np.eye(order, dtype=np.longdouble)

    # forward Gauss elimination with full pivoting:
    for k in range(order):
        # find pivot:
        im, jm = np.unravel_index(np.argmax(np.abs(m[k:, k:])),
                                  (order - k, order - k))
        im += k
        jm += k

        pv = m[im, jm]

        # detect singularity:
        if np.abs(pv) < eps:
            raise np.linalg.LinAlgError('Singular matrix.')

        if im != k or jm != k:
            # swap rows & columns:
            q = np.eye(order, dtype=int)

            # swap rows in the m and invm matrix:
            m[k, :], m[im, :] = m[im, :], m[k, :].copy()
            invm[k, :], invm[im, :] = invm[im, :], invm[k, :].copy()

            # swap rows in the Q matrix:
            q[k, :], q[jm, :] = q[jm, :], q[k, :].copy()

            # swap columns:
            m = np.dot(m, q)
            invm = np.dot(invm, q)
            qt = np.dot(qt, q)

        m[k, k:] /= pv
        r = m[k, (k + 1):]

        invm[k, :] /= pv
        w = invm[k, :]

        for l in range(k + 1, order):  # noqa: E741
            pv2 = m[l, k]
            m[l, (k + 1):] -= pv2 * r
            m[l, k] = 0.0
            invm[l, :] -= pv2 * w

    # inverse Jordan elimination:
    for k1 in range(order - 1, 0, -1):
        for k2 in range(k1 - 1, -1, -1):
            invm[k2, :] -= m[k2, k1] * invm[k1, :]

    # detect singularity:
    if not np.all(np.isfinite(invm)):
        raise np.linalg.LinAlgError('Singular matrix.')

    # undo permutations:
    invm = np.dot(qt, np.dot(invm, qt.T))

    return invm



"""
A module containing unit tests for the `wcsutil` module.

Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
import math
from distutils.version import LooseVersion

import numpy as np
from astropy.modeling import Model, Parameter
from astropy.modeling.models import AffineTransformation2D, Identity
from astropy import coordinates as coord
from astropy import units as u

try:
    import gwcs
    if LooseVersion(gwcs.__version__) > '0.12.0':
        from gwcs.geometry import CartesianToSpherical, SphericalToCartesian
        _GWCS_VER_GT_0P12 = True
    else:
        _GWCS_VER_GT_0P12 = False
except (ImportError, ModuleNotFoundError):
    _GWCS_VER_GT_0P12 = False

if _GWCS_VER_GT_0P12:
    _S2C = SphericalToCartesian(name='s2c', wrap_lon_at=180)
    _C2S = CartesianToSpherical(name='c2s', wrap_lon_at=180)

else:
    def _S2C(phi, theta):
        phi = np.deg2rad(phi)
        theta = np.deg2rad(theta)
        cs = np.cos(theta)
        x = cs * np.cos(phi)
        y = cs * np.sin(phi)
        z = np.sin(theta)
        return x, y, z

    def _C2S(x, y, z):
        h = np.hypot(x, y)
        phi = np.rad2deg(np.arctan2(y, x))
        theta = np.rad2deg(np.arctan2(z, h))
        return phi, theta


from tweakwcs.tpwcs import TPWCS, JWSTgWCS


class DummyTPWCS(TPWCS):
    def set_correction(self, matrix=[[1, 0], [0, 1]], shift=[0, 0], meta=None,
                       **kwargs):
        super().set_correction(matrix=matrix, shift=shift, meta=meta, **kwargs)

    def det_to_world(self, x, y):
        return super().det_to_world(x, y)

    def world_to_det(self, ra, dec):
        return super().world_to_det(ra, dec)

    def det_to_tanp(self, x, y):
        return super().det_to_tanp(x, y)

    def tanp_to_det(self, x, y):
        return super().tanp_to_det(x, y)

    def world_to_tanp(self, ra, dec):
        return super().world_to_tanp(ra, dec)

    def tanp_to_world(self, x, y):
        return super().tanp_to_world(x, y)


def rot_mat3d(angle, axis):
    cs = math.cos(angle)
    sn = math.sin(angle)
    axisv = np.array(axis * [0.0] + [1.0] + (2 - axis) * [0.0],
                     dtype=np.double)
    mat2d = np.array([[cs, sn], [-sn, cs]], dtype=np.double)
    return np.insert(np.insert(mat2d, axis, [0.0, 0.0], 1), axis, axisv, 0)


def create_DetToV2V3(v2ref=0.0, v3ref=0.0, roll=0.0,
                     cd=[[1.0, 0.0], [0.0, 1.0]], crpix=[0, 0]):
    tpcorr = JWSTgWCS._tpcorr_init(v2_ref=v2ref, v3_ref=v3ref, roll_ref=roll)

    afinv = AffineTransformation2D(cd, -np.dot(cd, crpix)).inverse

    JWSTgWCS._tpcorr_combine_affines(
        tpcorr,
        afinv.matrix.value,
        afinv.translation.value
    )

    p = JWSTgWCS._v2v3_to_tpcorr_from_full(tpcorr)
    partial_tpcorr = p.inverse
    partial_tpcorr.inverse = p

    return partial_tpcorr


def create_V2V3ToDet(v2ref=0.0, v3ref=0.0, roll=0.0,
                     cd=[[1.0, 0.0], [0.0, 1.0]], crpix=[0, 0]):
    inv_partial_tpcorr = create_DetToV2V3(
        v2ref=v2ref, v3ref=v3ref, roll=roll, crpix=crpix, cd=cd
    ).inverse
    return inv_partial_tpcorr


class V2V3ToSky(Model):
    """
    Rotates V2-V3 sphere on the sky.
    """
    angles = Parameter()

    _separable = False
    standard_broadcasting = False

    n_inputs = 2
    n_outputs = 2

    def __init__(self, angles, axes_order, name=None):
        self.axes_order = axes_order
        super().__init__(angles=angles, name=name)
        self.inputs = ("v2", "v3")
        self.outputs = ("ra", "dec")

    @staticmethod
    def build_euler_matrix(axis_angle):
        # build Euler rotation matrices:
        rotm = [rot_mat3d(np.deg2rad(alpha), axis)
                for axis, alpha in axis_angle]
        euler_rot = np.linalg.multi_dot(rotm)
        return euler_rot

    @staticmethod
    def cartesian2spherical(x, y, z):
        """ Convert cartesian coordinates to spherical (in deg). """
        return _C2S(x, y, z)

    @staticmethod
    def spherical2cartesian(alpha, delta):
        """ Convert spherical coordinates (in deg) to cartesian. """
        return _S2C(alpha, delta)

    def evaluate(self, v2, v3, angles):
        """ Evaluate the model on some input variables. """

        # convert spherical coordinates to cartesian assuming unit sphere:
        xyz = self.spherical2cartesian(v2.ravel() / 3600., v3.ravel() / 3600.0)

        # build Euler rotation matrices:
        euler_rot = self.__class__.build_euler_matrix(
            [(v[0], -v[1]) if v[0] != 1 else v for v in
             zip(self.axes_order, angles[0])][::-1]
        )

        # rotate cartezian coordinates:
        z, x, y = np.dot(euler_rot, xyz)

        # convert cartesian to spherical coordinates:
        ra, dec = self.cartesian2spherical(z, x, y)

        return ra, dec

    @property
    def inverse(self):
        return V2V3ToSkyInv(self.angles.value, self.axes_order)


class V2V3ToSkyInv(Model):
    """
    Rotates V2-V3 sphere on the sky.
    """
    angles = Parameter()

    _separable = False
    standard_broadcasting = False

    n_inputs = 2
    n_outputs = 2

    def __init__(self, angles, axes_order, name=None):
        self.axes_order = axes_order
        super().__init__(angles=angles, name=name)
        self.inputs = ("v2", "v3")
        self.outputs = ("ra", "dec")

    @staticmethod
    def build_euler_matrix(axis_angle):
        # build Euler rotation matrices:
        rotm = [rot_mat3d(np.deg2rad(alpha), axis)
                for axis, alpha in axis_angle]
        euler_rot = np.linalg.multi_dot(rotm)
        return euler_rot

    @staticmethod
    def cartesian2spherical(x, y, z):
        """ Convert cartesian coordinates to spherical (in deg). """
        return _C2S(x, y, z)

    @staticmethod
    def spherical2cartesian(alpha, delta):
        """ Convert spherical coordinates (in deg) to cartesian. """
        return _S2C(alpha, delta)

    def evaluate(self, v2, v3, angles):
        """ Evaluate the model on some input variables. """

        # convert spherical coordinates to cartesian assuming unit sphere:
        xyz = self.spherical2cartesian(v2.ravel(), v3.ravel())

        # build Euler rotation matrices:
        euler_rot = self.__class__.build_euler_matrix(
            [(v[0], -v[1]) if v[0] != 1 else v for v in
             zip(self.axes_order[::-1], (-angles[0])[::-1])][::-1]
        )

        # rotate cartezian coordinates:
        z, x, y = np.dot(euler_rot, xyz)

        # convert cartesian to spherical coordinates:
        ra, dec = self.cartesian2spherical(z, x, y)

        return 3600 * ra, 3600 * dec

    @property
    def inverse(self):
        return V2V3ToSky(self.angles.value, self.axes_order)


def make_mock_jwst_pipeline(v2ref=0, v3ref=0, roll=0, crpix=[512, 512],
                            cd=[[1e-5, 0], [0, 1e-5]], crval=[0, 0],
                            enable_vacorr=True):
    detector = gwcs.coordinate_frames.Frame2D(
        name='detector', axes_order=(0, 1), unit=(u.pix, u.pix)
    )
    v2v3 = gwcs.coordinate_frames.Frame2D(
        name='v2v3', axes_order=(0, 1), unit=(u.arcsec, u.arcsec)
    )
    v2v3vacorr = gwcs.coordinate_frames.Frame2D(
        name='v2v3vacorr', axes_order=(0, 1), unit=(u.arcsec, u.arcsec)
    )
    world = gwcs.coordinate_frames.CelestialFrame(reference_frame=coord.ICRS(),
                                                  name='world')
    det2v2v3 = create_DetToV2V3(v2ref=v2ref / 3600.0, v3ref=v3ref / 3600.0,
                                roll=roll, cd=cd, crpix=crpix)

    v23sky = V2V3ToSky([-v2ref / 3600.0, v3ref / 3600.0, -roll,
                        -crval[1], crval[0]], [2, 1, 0, 1, 2])
    if enable_vacorr:
        pipeline = [(detector, det2v2v3), (v2v3, Identity(2)),
                    (v2v3vacorr, v23sky), (world, None)]
    else:
        pipeline = [(detector, det2v2v3), (v2v3, v23sky), (world, None)]

    return pipeline


def make_mock_jwst_wcs(v2ref=0, v3ref=0, roll=0, crpix=[512, 512],
                       cd=[[1e-5, 0], [0, 1e-5]], crval=[0, 0],
                       enable_vacorr=True):
    pipeline = make_mock_jwst_pipeline(v2ref, v3ref, roll, crpix, cd, crval, enable_vacorr)
    wcs = gwcs.wcs.WCS(pipeline)
    wcs.bounding_box = ((-0.5, 1024 - 0.5), (-0.5, 2048 - 0.5))
    wcs.array_shape = (2048, 1024)
    return wcs



"""
A module containing unit tests for the `tpwcs` module.

Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
import copy
import pytest
from distutils.version import LooseVersion

import numpy as np
from astropy.modeling.models import Scale, Identity
from astropy import wcs as fitswcs

try:
    import gwcs
    if LooseVersion(gwcs.__version__) > '0.12.0':
        from gwcs.geometry import SphericalToCartesian
        _GWCS_VER_GT_0P12 = True
    else:
        _GWCS_VER_GT_0P12 = False
except ImportError:
    _GWCS_VER_GT_0P12 = False

import astropy
if LooseVersion(astropy.__version__) >= '4.0':
    _ASTROPY_VER_GE_4 = True
    from astropy.modeling import CompoundModel
else:
    _ASTROPY_VER_GE_4 = False

from tweakwcs.linearfit import build_fit_matrix
from tweakwcs import tpwcs

from .helper_tpwcs import (make_mock_jwst_wcs, make_mock_jwst_pipeline,
                           DummyTPWCS, create_DetToV2V3, create_V2V3ToDet)


_ATOL = 100 * np.finfo(np.array([1.]).dtype).eps
_NO_JWST_SUPPORT = not (_ASTROPY_VER_GE_4 and _GWCS_VER_GT_0P12)


def test_tpwcs():
    tpwcs = DummyTPWCS(None, meta={})
    x, y, ra, dec = np.random.random(4)
    matrix = np.random.random((2, 2))
    shift = np.random.random(2)

    assert tpwcs.world_to_det(ra, dec) == (ra, dec)
    assert tpwcs.world_to_tanp(ra, dec) == (ra, dec)
    assert tpwcs.det_to_world(x, y) == (x, y)
    assert tpwcs.det_to_tanp(x, y) == (x, y)
    assert tpwcs.tanp_to_det(x, y) == (x, y)
    assert tpwcs.tanp_to_world(x, y) == (x, y)
    assert tpwcs.tanp_center_pixel_scale == 1
    assert tpwcs.wcs is None
    assert tpwcs.original_wcs is None
    assert isinstance(tpwcs.copy(), DummyTPWCS)
    assert tpwcs.bounding_box is None

    tpwcs.set_correction(matrix=matrix, shift=shift,
                         meta={'pytest': 'ABC.TPWCS'}, pytest_kwarg=True)
    assert np.all(tpwcs.meta['matrix'] == matrix)
    assert np.all(tpwcs.meta['shift'] == shift)
    assert tpwcs.meta['pytest'] == 'ABC.TPWCS'

    with pytest.raises(TypeError) as arg_err:
        tpwcs.set_correction(matrix, shift, {'pytest': 'ABC.TPWCS'},
                             'some_weird_arg')
    assert (arg_err.value.args[0] == "set_correction() takes from 1 "
            "to 4 positional arguments but 5 were given")


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_mock_jwst_gwcs():
    w = make_mock_jwst_wcs(v2ref=123, v3ref=500, roll=115, crpix=[-512, -512],
                           cd=[[1e-5, 0], [0, 1e-5]], crval=[82, 12])

    assert np.allclose(w.invert(*w(23, 1023)), (23, 1023))


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
@pytest.mark.parametrize('crpix, cd', [
    (np.zeros(3), np.diag(np.ones(3))),
    (np.zeros((2, 2)), np.diag(np.ones(2))),
])
def test_mock_wcs_fails(crpix, cd):
    from astropy.modeling import InputParameterError
    with pytest.raises(InputParameterError):
        make_mock_jwst_wcs(v2ref=123, v3ref=500, roll=15, crpix=crpix,
                           cd=cd, crval=[82, 12])
    with pytest.raises(InputParameterError):
        create_DetToV2V3(v2ref=123, v3ref=500, roll=15, crpix=crpix, cd=cd)
    with pytest.raises(InputParameterError):
        create_V2V3ToDet(v2ref=123, v3ref=500, roll=15, crpix=crpix, cd=cd)


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_v2v3todet_roundtrips():
    s2c = (Scale(1.0 / 3600.0) & Scale(1.0 / 3600.0)) | SphericalToCartesian(wrap_lon_at=180)

    s = 1.0e-5
    crpix = np.random.random(2)
    alpha = 0.25 * np.pi * np.random.random()
    x, y = 1024 * np.random.random(2)
    v2, v3 = 45 * np.random.random(2)
    cd = [[s * np.cos(alpha), -s * np.sin(alpha)],
          [s * np.sin(alpha), s * np.cos(alpha)]]

    d2v = create_DetToV2V3(
        v2ref=123.0, v3ref=500.0, roll=15.0, crpix=crpix, cd=cd
    )
    v2d = create_V2V3ToDet(
        v2ref=123.0, v3ref=500.0, roll=15.0, crpix=crpix, cd=cd
    )

    assert np.allclose(d2v.inverse(*d2v(x, y)), (x, y),
                       rtol=1e3 * _ATOL, atol=1e3 * _ATOL)

    assert (
        np.allclose(
            s2c(*v2d.inverse(*v2d(v2, v3))),
            s2c(v2, v3),
            rtol=1e3 * _ATOL, atol=_ATOL
        ) or np.allclose(
            -np.asanyarray(s2c(*v2d.inverse(*v2d(v2, v3)))),
            s2c(v2, v3),
            rtol=1e3 * _ATOL, atol=_ATOL
        )
    )
    assert np.allclose(v2d(*d2v(x, y)), (x, y),
                       rtol=1e5 * _ATOL, atol=1e3 * _ATOL)

    assert (
        np.allclose(
            s2c(*d2v(*v2d(v2, v3))),
            s2c(v2, v3),
            rtol=1e3 * _ATOL, atol=1e3 * _ATOL
        ) or np.allclose(
            -np.asanyarray(s2c(*d2v(*v2d(v2, v3)))),
            s2c(v2, v3),
            rtol=1e3 * _ATOL, atol=1e3 * _ATOL
        )
    )


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_jwst_wcs_corr_applied(mock_jwst_wcs):
    w = make_mock_jwst_wcs(
        v2ref=123.0, v3ref=500.0, roll=115.0, crpix=[512.0, 512.0],
        cd=[[1.0e-5, 0.0], [0.0, 1.0e-5]], crval=[82.0, 12.0]
    )

    wc = tpwcs.JWSTgWCS(
        w, {'v2_ref': 123.0, 'v3_ref': 500.0, 'roll_ref': 115.0}, meta={}
    )
    wc.set_correction(meta={'dummy_meta': None}, dummy_par=1)
    assert 'v2v3corr' in wc.wcs.available_frames
    assert 'dummy_meta' in wc.meta


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_jwst_wcs_corr_are_being_combined(mock_jwst_wcs):
    wc = tpwcs.JWSTgWCS(
        mock_jwst_wcs, {'v2_ref': 123.0, 'v3_ref': 500.0, 'roll_ref': 115.0}
    )
    matrix1 = np.array([[1.0, 0.2], [-0.3, 1.1]])
    shift1 = np.array([5.0, -7.0])
    wc.set_correction(matrix=matrix1, shift=shift1)
    assert 'v2v3corr' in wc.wcs.available_frames

    matrix2 = np.linalg.inv(matrix1)
    shift2 = -np.dot(matrix2, shift1)
    wc = tpwcs.JWSTgWCS(
        wc.wcs, {'v2_ref': 123.0, 'v3_ref': 500.0, 'roll_ref': 115.0}
    )
    wc.set_correction(matrix=matrix2, shift=shift2)

    v2v3idx = [k for k, n in enumerate(wc.wcs.available_frames)
               if n == 'v2v3corr']

    assert len(v2v3idx) == 1

    tp_corr = wc.wcs.pipeline[v2v3idx[0] - 1].transform

    assert isinstance(tp_corr, CompoundModel)
    assert np.max(np.abs(tp_corr['tp_affine'].matrix.value - np.identity(2))) < _ATOL
    assert np.max(np.abs(tp_corr['tp_affine'].translation.value)) < _ATOL


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_jwstgwcs_unsupported_wcs():
    from tweakwcs import tpwcs
    dummy_wcs = gwcs.wcs.WCS(Identity(2), 'det', 'world')
    with pytest.raises(ValueError):
        tpwcs.JWSTgWCS(dummy_wcs, {})


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_jwstgwcs_inconsistent_ref(mock_jwst_wcs):
    wc = tpwcs.JWSTgWCS(
        mock_jwst_wcs, {'v2_ref': 123.0, 'v3_ref': 500.0, 'roll_ref': 115.0},
    )
    wc.set_correction()

    with pytest.raises(ValueError):
        wc = tpwcs.JWSTgWCS(
            wc.wcs, {'v2_ref': 124.0, 'v3_ref': 500.0, 'roll_ref': 115.0},
        )


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_jwstgwcs_wrong_tpcorr_type(mock_jwst_wcs):
    wc = tpwcs.JWSTgWCS(
        mock_jwst_wcs, {'v2_ref': 123.0, 'v3_ref': 500.0, 'roll_ref': 115.0},
    )
    wc.set_correction()
    p = wc.wcs.pipeline

    np = [(v[0], create_V2V3ToDet()) if v[0].name == 'v2v3vacorr'
          else v for v in p]
    mangled_wc = gwcs.wcs.WCS(np)

    with pytest.raises(ValueError):
        wc = tpwcs.JWSTgWCS(
            mangled_wc, {'v2_ref': 123.0, 'v3_ref': 500.0, 'roll_ref': 115.0},
        )


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_jwstgwcs_ref_angles_preserved(mock_jwst_wcs):
    wc = tpwcs.JWSTgWCS(
        mock_jwst_wcs, {'v2_ref': 123.0, 'v3_ref': 500.0, 'roll_ref': 115.0},
    )
    assert wc.ref_angles['v2_ref'] == 123.0
    assert wc.ref_angles['v3_ref'] == 500.0
    assert wc.ref_angles['roll_ref'] == 115.0

# Test inputs with different shapes


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
@pytest.mark.parametrize('inputs', [[500, 512, 12, 24],
                                    [[500, 500], [512, 512], [12, 12], [24, 24]],
                                    [[[500, 500], [500, 500]],
                                     [[512, 512], [512, 512]],
                                     [[12, 12], [12, 12]], [[24, 24], [24, 24]]]
                                    ])
def test_jwstgwcs_detector_to_world(inputs):
    w = make_mock_jwst_wcs(
        v2ref=0.0, v3ref=0.0, roll=0.0, crpix=[500.0, 512.0],
        cd=[[1.0e-5, 0.0], [0.0, 1.0e-5]], crval=[12.0, 24.0]
    )
    wc = tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})
    wc.set_correction()
    x, y, ra, dec = inputs
    assert np.allclose(wc.det_to_world(x, y), (ra, dec), atol=_ATOL)
    assert np.allclose(wc.world_to_det(ra, dec), (x, y), atol=_ATOL)


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
@pytest.mark.parametrize('inputs', [[0, 0, 12, 24],
                                    [[0, 0], [0, 0], [12, 12], [24, 24]],
                                    [[[0, 0], [0, 0]],
                                     [[0, 0], [0, 0]],
                                     [[12, 12], [12, 12]], [[24, 24], [24, 24]]]
                                    ])
def test_jwstgwcs_tangent_to_world(inputs):
    w = make_mock_jwst_wcs(
        v2ref=0.0, v3ref=0.0, roll=0.0, crpix=[500.0, 512.0],
        cd=[[1.0e-5, 0.0], [0.0, 1.0e-5]], crval=[12.0, 24.0]
    )
    wc = tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})
    wc.set_correction()
    tanp_x, tanp_y, ra, dec = inputs
    assert np.allclose(wc.world_to_tanp(ra, dec), (tanp_x, tanp_y), atol=1000 * _ATOL)
    assert np.allclose(wc.tanp_to_world(tanp_x, tanp_y), (ra, dec), atol=_ATOL)


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
@pytest.mark.parametrize('inputs', [[500, 512, 0, 0],
                                    [[500, 500], [512, 512], [0, 0], [0, 0]],
                                    [[[500, 500], [500, 500]],
                                     [[512, 512], [512, 512]],
                                     [[0, 0], [0, 0]], [[0, 0], [0, 0]]]
                                    ])
def test_jwstgwcs_detector_to_tanp(inputs):
    w = make_mock_jwst_wcs(
        v2ref=0.0, v3ref=0.0, roll=0.0, crpix=[500.0, 512.0],
        cd=[[1.0e-5, 0.0], [0.0, 1.0e-5]], crval=[12.0, 24.0]
    )
    wc = tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})
    wc.set_correction()
    x, y, tanp_x, tanp_y = inputs
    assert np.allclose(wc.det_to_tanp(x, y), (tanp_x, tanp_y), atol=10 * _ATOL)
    assert np.allclose(wc.tanp_to_det(tanp_x, tanp_y), (x, y), atol=_ATOL)


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_jwstgwcs_bbox():
    w = make_mock_jwst_wcs(
        v2ref=0.0, v3ref=0.0, roll=0.0, crpix=[500.0, 512.0],
        cd=[[1.0e-5, 0.0], [0.0, 1.0e-5]], crval=[12.0, 24.0]
    )
    wc = tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})
    wc.set_correction()

    assert np.allclose(
        wc.bounding_box,
        ((-0.5, 1024 - 0.5), (-0.5, 2048 - 0.5)),
        atol=_ATOL
    )

    wc._owcs.bounding_box = None
    assert np.allclose(
        wc.bounding_box,
        ((-0.5, 1024 - 0.5), (-0.5, 2048 - 0.5)),
        atol=_ATOL
    )

    wc._owcs.array_shape = None
    assert wc.bounding_box is None


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_jwstgwcs_bad_pipelines_no_vacorr():
    p0 = make_mock_jwst_pipeline(
        v2ref=0.0, v3ref=0.0, roll=0.0, crpix=[500.0, 512.0],
        cd=[[1.0e-5, 0.0], [0.0, 1.0e-5]], crval=[12.0, 24.0],
        enable_vacorr=False
    )

    # no pipeline or empty pipeline:
    with pytest.raises(ValueError):
        tpwcs.JWSTgWCS(None, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})

    # fewer than 3 frames:
    w = gwcs.wcs.WCS(p0[:2])
    with pytest.raises(ValueError):
        tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})

    # repeated (any one of the) last two frames:
    w = gwcs.wcs.WCS(p0 + [p0[-1]])
    with pytest.raises(ValueError):
        tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})

    w = gwcs.wcs.WCS(p0 + [p0[-2]])
    with pytest.raises(ValueError):
        tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})

    # multiple 'v2v3' frames:
    w = gwcs.wcs.WCS(p0)
    w = tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})
    p = w.wcs.pipeline
    p.insert(1, p[1])
    w = gwcs.wcs.WCS(p)
    with pytest.raises(ValueError):
        tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})

    # misplaced 'v2v3' frame:
    w = gwcs.wcs.WCS(p0)
    w = tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})
    w.set_correction()
    p = w.wcs.pipeline
    del p[0]
    w = gwcs.wcs.WCS(p)
    with pytest.raises(ValueError):
        tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})

    # multiple 'v2v3corr' frame:
    w = gwcs.wcs.WCS(p0)
    w = tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})
    w.set_correction()
    p = w.wcs.pipeline
    p.insert(1, p[-2])
    w = gwcs.wcs.WCS(p)
    with pytest.raises(ValueError):
        tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})

    # misplaced 'v2v3corr' frame:
    del p[-2]
    w = gwcs.wcs.WCS(p)
    with pytest.raises(ValueError):
        tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})


@pytest.mark.skipif(_NO_JWST_SUPPORT, reason="requires gwcs>=0.12.1")
def test_jwstgwcs_bad_pipelines_with_vacorr():
    p0 = make_mock_jwst_pipeline(
        v2ref=0.0, v3ref=0.0, roll=0.0, crpix=[500.0, 512.0],
        cd=[[1.0e-5, 0.0], [0.0, 1.0e-5]], crval=[12.0, 24.0],
        enable_vacorr=True
    )

    # multiple 'v2v3vacorr' frames:
    w = gwcs.wcs.WCS(p0)
    w = tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})
    p = w.wcs.pipeline
    p.insert(1, p[2])
    w = gwcs.wcs.WCS(p)
    with pytest.raises(ValueError):
        tpwcs.JWSTgWCS(w, {'v2_ref': 0.0, 'v3_ref': 0.0, 'roll_ref': 0.0})


def test_fitswcs_non_celestial():
    # non-celestial WCS
    w = fitswcs.WCS(naxis=3)
    with pytest.raises(ValueError):
        tpwcs.FITSWCS(w)

    # invalid WCS
    with pytest.raises(ValueError):
        tpwcs.FITSWCS(None)


def test_fitswcs_unaccounted_dist(mock_fits_wcs):
    w = copy.deepcopy(mock_fits_wcs)
    w.pix2foc = lambda x, o: x + 3
    with pytest.raises(ValueError):
        tpwcs.FITSWCS(w)

    w = copy.deepcopy(mock_fits_wcs)
    f = w.all_world2pix
    w.all_world2pix = lambda x, o: f(x, o) + 2
    with pytest.raises(ValueError):
        tpwcs.FITSWCS(w)


def test_fitswcs_1(mock_fits_wcs):
    wc = tpwcs.FITSWCS(mock_fits_wcs)
    wc.set_correction()


def test_fitswcs_coord_transforms(mock_fits_wcs):
    w = fitswcs.WCS(naxis=2)
    w.wcs.cd = build_fit_matrix(0, 1e-5)
    w.wcs.crval = [12, 24]
    w.wcs.crpix = [500, 512]
    w.wcs.ctype = ['RA---TAN', 'DEC--TAN']
    w.pixel_shape = [1024, 2048]
    w.wcs.set()

    wc = tpwcs.FITSWCS(w)
    wc.set_correction()

    assert np.allclose(wc.det_to_world(499, 511), (12, 24), atol=_ATOL)
    assert np.allclose(wc.world_to_det(12, 24), (499, 511), atol=_ATOL)
    assert np.allclose(wc.det_to_tanp(499, 511), (499, 511), atol=_ATOL)
    assert np.allclose(wc.tanp_to_det(499, 511), (499, 511), atol=_ATOL)
    assert np.allclose(wc.world_to_tanp(12, 24), (499, 511), atol=1e-8)
    assert np.allclose(wc.tanp_to_world(499, 511), (12, 24), atol=_ATOL)


def test_fitswcs_bbox(mock_fits_wcs):
    w = copy.deepcopy(mock_fits_wcs)
    wc = tpwcs.FITSWCS(w)
    wc.set_correction()

    assert np.allclose(
        wc.bounding_box,
        ((-0.5, 1024 - 0.5), (-0.5, 2048 - 0.5)),
        atol=_ATOL
    )

    wc._owcs.pixel_bounds = None
    assert np.allclose(
        wc.bounding_box,
        ((-0.5, 1024 - 0.5), (-0.5, 2048 - 0.5)),
        atol=_ATOL
    )

    wc._owcs.bounding_box = None
    assert np.allclose(
        wc.bounding_box,
        ((-0.5, 1024 - 0.5), (-0.5, 2048 - 0.5)),
        atol=_ATOL
    )

    wc._owcs.array_shape = None
    assert wc.bounding_box is None



"""
A module containing unit tests for the `wcsimage` module.

Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
import copy
import pytest
import numpy as np
from astropy.table import Table, Column
from tweakwcs import TPMatch, FITSWCS
from tweakwcs.wcsimage import (convex_hull, RefCatalog, WCSImageCatalog,
                               WCSGroupCatalog, _is_int)


_ATOL = 100 * np.finfo(np.array([1.]).dtype).eps


@pytest.fixture
def rect_cat(scope='function'):
    x = np.array([0.0, 0.0, 10.0, 10.0, 0.0]) - 5
    y = np.array([0.0, 10.0, 10.0, 0.0, 0.0]) - 5
    imcat = Table([x, y], names=('x', 'y'))
    return imcat


@pytest.fixture(scope='function')
def rect_imcat(mock_fits_wcs):
    x = np.array([0.0, 0.0, 10.0, 10.0, 0.0]) - 5
    y = np.array([0.0, 10.0, 10.0, 0.0, 0.0]) - 5
    imcat = Table([x, y], names=('x', 'y'))
    tpwcs = FITSWCS(mock_fits_wcs)
    w = WCSImageCatalog(imcat, tpwcs)
    return w


@pytest.mark.parametrize('val, expected', [
    (1, True), (1.0, False), ('1', False), (True, False),
    (np.bool_(1), False), (np.float16(1.0), False), (np.int16(1), True),
])
def test_is_int(val, expected):
    assert _is_int(val) is expected


def test_wcsimcat_wrong_tpwcs_type(mock_fits_wcs):
    imcat = Table([[], []], names=('x', 'y'))
    with pytest.raises(TypeError) as e:
        WCSImageCatalog(imcat, None)
    assert (e.value.args[0] == "Unsupported 'tpwcs' type. 'tpwcs' must be "
            "a subtype of TPWCS.")


def test_wcsimcat_wrong_fit_status(mock_fits_wcs, rect_imcat):
    rect_imcat.fit_status = 'spectacular'
    assert rect_imcat.fit_status == 'spectacular'


def test_wcsimcat_set_none_catalog(mock_fits_wcs, rect_imcat):
    rect_imcat.catalog = None
    assert rect_imcat.catalog is None


def test_wcsimcat_wcs_transforms_roundtrip(mock_fits_wcs):
    x = np.arange(5)
    y = np.random.permutation(5)
    imcat = Table([x, y], names=('x', 'y'))
    tpwcs = FITSWCS(mock_fits_wcs)
    w = WCSImageCatalog(imcat, tpwcs)

    assert np.allclose(
        w.world_to_det(*w.det_to_world(x, y)), (x, y), rtol=0, atol=1.0e-5
    )

    assert np.allclose(
        w.tanp_to_det(*w.det_to_tanp(x, y)), (x, y), rtol=0, atol=1.0e-5
    )

    assert np.allclose(
        w.world_to_tanp(*w.tanp_to_world(1.0e-5 * x, 1.0e-5 * y)),
        (1.0e-5 * x, 1.0e-5 * y),
        rtol=0, atol=1.0e-5
    )


def test_wcsimcat_intersections(mock_fits_wcs, rect_imcat):
    pts1 = list(rect_imcat.polygon.points)[0]
    pts2 = list(rect_imcat.intersection(rect_imcat.polygon).points)[0]
    for pt1 in pts1:
        assert any(np.allclose(pt1, pt2) for pt2 in pts2)

    pts2 = list(rect_imcat.intersection(rect_imcat).points)[0]
    for pt1 in pts1:
        assert any(np.allclose(pt1, pt2) for pt2 in pts2)

    assert np.allclose(
        rect_imcat.intersection_area(rect_imcat), 2.9904967391303217e-12,
        atol=0.0, rtol=5.0e-4
    )


def test_wcsimcat_guarded_intersection_area(mock_fits_wcs, rect_imcat):
    assert np.allclose(
        rect_imcat._guarded_intersection_area(rect_imcat),
        2.9904967391303217e-12, atol=0.0, rtol=5.0e-4
    )


def test_wcsimcat_no_wcs_bb(mock_fits_wcs, rect_cat):
    tpwcs = FITSWCS(mock_fits_wcs)
    tpwcs._owcs.pixel_bounds = None
    tpwcs._owcs.pixel_shape = None

    assert tpwcs.bounding_box is None

    WCSImageCatalog(rect_cat, tpwcs)


def test_wcsimcat_calc_chip_bounding_polygon_custom_stepsize(mock_fits_wcs,
                                                             rect_imcat):
    rect_imcat._calc_chip_bounding_polygon(stepsize=None)
    pts1 = list(rect_imcat.polygon.points)[0]

    rect_imcat._calc_chip_bounding_polygon(stepsize=2)
    pts2 = list(rect_imcat.polygon.points)[0]

    for pt1 in pts1:
        assert any(np.allclose(pt1, pt2) for pt2 in pts2)

    assert rect_imcat._calc_chip_bounding_polygon


def test_wcsimcat_calc_cat_convex_hull_no_catalog(mock_fits_wcs, rect_imcat):
    rect_imcat.catalog = None
    assert rect_imcat._calc_cat_convex_hull() is None


def test_wcsimcat_bb_radec(mock_fits_wcs, rect_imcat):
    ra1, dec1 = rect_imcat.det_to_world(rect_imcat.catalog['x'],
                                        rect_imcat.catalog['y'])
    ra2, dec2 = rect_imcat.bb_radec
    assert np.allclose(ra1, ra2[::-1])
    assert np.allclose(dec1, dec2[::-1])


def test_wcsgroupcat_init(mock_fits_wcs, rect_imcat):
    g = WCSGroupCatalog(rect_imcat, 'name1')
    assert g.name == 'name1'

    for im in g:
        assert im.name == 'Unknown'

    g.name = 'name2'
    assert g.name == 'name2'
    assert g[0] is rect_imcat

    assert len(g) == 1

    # input list is empty:
    with pytest.raises(ValueError) as e:
        WCSGroupCatalog([])
    assert e.value.args[0] == "List of images cannot be empty."

    # wrong type for the WCSImageCatalog in a list
    with pytest.raises(TypeError) as e:
        WCSGroupCatalog([1])
    assert (e.value.args[0] == "Each element of the 'images' parameter "
            "must be an 'WCSImageCatalog' object.")

    # wrong type for the input catalog: it is not a WCSImageCatalog
    with pytest.raises(TypeError) as e:
        WCSGroupCatalog(1)
    assert (e.value.args[0] == "Parameter 'images' must be either a single "
            "'WCSImageCatalog' object or a list of 'WCSImageCatalog' objects")

    # input WCSImageCatalog with a missing catalog:
    with pytest.raises(ValueError) as e:
        rect_imcat.catalog = None
        WCSGroupCatalog(rect_imcat)
    assert (e.value.args[0] == "Each input WCS image catalog must have a "
            "valid catalog.")

    # input WCSImageCatalog  with a missing catalog in the input list:
    with pytest.raises(ValueError) as e:
        rect_imcat.catalog = None
        WCSGroupCatalog([rect_imcat])
    assert (e.value.args[0] == "Each input WCS image catalog must have a "
            "valid catalog.")


def test_wcsgroupcat_intersections(mock_fits_wcs, rect_imcat):
    g = WCSGroupCatalog(rect_imcat)

    pts1 = list(g.polygon.points)[0]
    pts2 = list(g.intersection(g.polygon).points)[0]
    for pt1 in pts1:
        assert any(np.allclose(pt1, pt2) for pt2 in pts2)

    pts2 = list(g.intersection(g).points)[0]
    for pt1 in pts1:
        assert any(np.allclose(pt1, pt2) for pt2 in pts2)

    assert np.allclose(
        g.intersection_area(g), 2.9904967391303217e-12, atol=0.0, rtol=5.0e-4
    )


def test_wcsgroupcat_guarded_intersection_area(mock_fits_wcs, rect_imcat):
    g = WCSGroupCatalog(rect_imcat)
    assert np.allclose(
        g._guarded_intersection_area(g), 2.9904967391303217e-12,
        atol=0.0, rtol=5.0e-4
    )


def test_wcsgroupcat_update_bb_no_images(mock_fits_wcs, rect_imcat):
    g = WCSGroupCatalog(rect_imcat)
    g._images = []
    g.update_bounding_polygon()
    assert len(g.polygon) == 0


def test_wcsgroupcat_create_group_catalog(mock_fits_wcs, rect_imcat):
    w1 = copy.deepcopy(rect_imcat)
    w2 = copy.deepcopy(rect_imcat)
    g = WCSGroupCatalog([w1, w2])

    # catalogs with name set to None:
    names = []
    for im in g:
        names.append(im.name)
        im._name = None
    assert len(g.create_group_catalog()) == 2 * len(rect_imcat.catalog)
    for im, name in zip(g, names):
        im.name = name

    # Mixed catalogs: one has weights and another does not:
    g[0].catalog.add_column(Column(np.ones(5)), name='weight')
    with pytest.raises(KeyError) as e:
        g.create_group_catalog()
        assert False
    assert (e.value.args[0] == "Non-empty catalogs in a group must all "
            "either have or not have 'weight' column.")
    g[0].catalog.remove_column('weight')

    # artificially set catalog of one of the images to an empty table:
    g[0].catalog.remove_rows(slice(None, None))
    assert len(g.create_group_catalog()) == len(rect_imcat.catalog)

    # artificially set all catalogs to empty table:
    for im in g:
        im.catalog.remove_rows(slice(None, None))
    assert not g.create_group_catalog()

    # artificially set all catalogs to empty table:
    for im in g:
        im._name = None
    assert not g.create_group_catalog()


def test_wcsgroupcat_recalc_catalog_radec(mock_fits_wcs, rect_imcat):
    ra, dec = mock_fits_wcs.all_pix2world(rect_imcat.catalog['x'],
                                          rect_imcat.catalog['y'], 0)
    refcat = Table([ra, dec], names=('RA', 'DEC'))
    ref = RefCatalog(refcat)

    wim1 = copy.deepcopy(rect_imcat)
    wim2 = copy.deepcopy(rect_imcat)
    g = WCSGroupCatalog(wim1)

    ra = g.catalog['RA']
    dec = g.catalog['DEC']

    # artificially add another image:
    g._images.append(wim2)

    g.recalc_catalog_radec()

    assert np.allclose(g.catalog['RA'], ra)
    assert np.allclose(g.catalog['DEC'], dec)

    g.align_to_ref(ref)
    mcat = g.get_matched_cat()
    assert len(mcat) == 5
    assert np.allclose(mcat['RA'], ra)
    assert np.allclose(mcat['DEC'], dec)

    del g.catalog['RA']
    del g.catalog['DEC']
    with pytest.raises(RuntimeError):
        g.calc_tanp_xy(None)


def test_wcsgroupcat_match2ref(mock_fits_wcs, rect_imcat):
    ra, dec = mock_fits_wcs.all_pix2world(rect_imcat.catalog['x'],
                                          rect_imcat.catalog['y'], 0)
    refcat = Table([ra[:-1], dec[:-1]], names=('RA', 'DEC'))
    ref = RefCatalog(refcat)

    # unequal catalog lengths
    g = WCSGroupCatalog(rect_imcat)
    with pytest.raises(ValueError):
        g.match2ref(ref, match=None)

    refcat = Table([ra, dec], names=('RA', 'DEC'))
    ref = RefCatalog(refcat)

    # call calc_tanp_xy before matching
    with pytest.raises(RuntimeError) as e:
        g.match2ref(ref, match=TPMatch())
    assert (e.value.args[0] == "'calc_tanp_xy()' should have been run "
            "prior to match2ref()")

    ref.calc_tanp_xy(rect_imcat.tpwcs)
    g.calc_tanp_xy(rect_imcat.tpwcs)
    g.catalog['matched_ref_id'] = np.ones(5, dtype=bool)
    g.catalog['_raw_matched_ref_idx'] = np.ones(5, dtype=bool)
    g.match2ref(ref, match=TPMatch())


def test_wcsgroupcat_fit2ref(mock_fits_wcs, caplog, rect_imcat):
    ra, dec = mock_fits_wcs.all_pix2world(rect_imcat.catalog['x'],
                                          rect_imcat.catalog['y'], 0)
    refcat = Table([ra, dec], names=('RA', 'DEC'))
    ref = RefCatalog(refcat)
    ref.calc_tanp_xy(rect_imcat.tpwcs)

    g = WCSGroupCatalog(rect_imcat)
    g.calc_tanp_xy(rect_imcat.tpwcs)
    g.match2ref(ref)

    g.fit2ref(ref, rect_imcat.tpwcs, fitgeom='shift')
    g.apply_affine_to_wcs(g[0].tpwcs, np.identity(2), np.zeros(2))

    g._images = []
    g.align_to_ref(ref)
    assert caplog.record_tuples[-1][-1].endswith("Nothing to align.")


def test_wcsrefcat_init_check_catalog(mock_fits_wcs):
    r = RefCatalog(None, name='refcat')

    assert r.name == 'refcat'
    r.name = 'refcat2'
    assert r.name == 'refcat2'

    assert r.poly_area is None

    with pytest.raises(ValueError) as e:
        r.catalog = None
    assert e.value.args[0] == "Reference catalogs cannot be None"

    with pytest.raises(KeyError) as e:
        r.catalog = Table([[1], [1]], names=('x', 'DEC'))
    assert (e.value.args[0] == "Reference catalogs *must* contain *both* 'RA' "
            "and 'DEC' columns.")

    r.expand_catalog(Table([[1], [1]], names=('RA', 'DEC')))


def test_wcsrefcat_intersections(mock_fits_wcs, rect_imcat):
    ra, dec = mock_fits_wcs.all_pix2world(10 * rect_imcat.catalog['x'],
                                          10 * rect_imcat.catalog['y'], 0)
    refcat = Table([ra, dec], names=('RA', 'DEC'))
    ref = RefCatalog(refcat)
    ref.calc_tanp_xy(rect_imcat.tpwcs)

    pts1 = list(ref.polygon.points)[0]
    pts2 = list(ref.intersection(ref.polygon).points)[0]
    for pt1 in pts1:
        assert any(np.allclose(pt1, pt2) for pt2 in pts2)

    pts2 = list(ref.intersection(ref).points)[0]
    for pt1 in pts1:
        assert any(np.allclose(pt1, pt2) for pt2 in pts2)

    assert np.allclose(
        ref.intersection_area(ref), 2.9902125220360176e-10, atol=0.0,
        rtol=0.005,
    )


def test_wcsrefcat_guarded_intersection_area(mock_fits_wcs, rect_imcat):
    ra, dec = mock_fits_wcs.all_pix2world(10 * rect_imcat.catalog['x'],
                                          10 * rect_imcat.catalog['y'], 0)
    refcat = Table([ra, dec], names=('RA', 'DEC'))
    ref = RefCatalog(refcat)
    ref.calc_tanp_xy(rect_imcat.tpwcs)

    assert np.allclose(
        ref._guarded_intersection_area(ref), 2.9902125220360176e-10, atol=0.0,
        rtol=0.005,
    )


def test_wcsrefcat_convex_hull(mock_fits_wcs):
    ref = RefCatalog(None)
    ref._calc_cat_convex_hull()  # catalog is None

    assert convex_hull([], []) == ([], [])
    assert all(isinstance(x, np.ndarray) and x.size == 0 for x in
               convex_hull(np.array([]), np.array([])))

    assert convex_hull([123], [456]) == ([123], [456])

    x = [0, 0, 10, 10]
    y = [0, 10, 10, 0]

    xc, yc = convex_hull(x + [10, 3, 6, 7], y + [0, 5, 1, 8])
    assert set(xc) == set(x) and set(yc) == set(y)

    xc, yc = convex_hull(np.array(x + [3, 6, 7]), np.array(y + [5, 1, 8]))
    assert set(xc) == set(x) and set(yc) == set(y)

    xc, yc = convex_hull(x + [3, 6, 7], y + [5, 1, 8], lambda x, y: (x, y))
    assert set(xc) == set(x) and set(yc) == set(y)

    xc, yc = convex_hull(np.array(x + [3, 6, 7]), np.array(y + [5, 1, 8]),
                         lambda x, y: (x, y))
    assert set(xc) == set(x) and set(yc) == set(y)



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

try:
    from distutils.config import ConfigParser
except ImportError:
    from configparser import ConfigParser

conf = ConfigParser()
conf.read(['setup.cfg'])

# Get some config values
metadata = dict(conf.items('metadata'))
PACKAGENAME = metadata.get('package_name', 'tweakwcs')
DESCRIPTION = metadata.get('description', 'A package for correcting alignment '
                           'errors in WCS objects')
LONG_DESCRIPTION = metadata.get('long_description', 'README.rst')
LONG_DESCRIPTION_CONTENT_TYPE = metadata.get('long_description_content_type',
                                             'text/x-rst')
AUTHOR = metadata.get('author', 'Mihai Cara')
AUTHOR_EMAIL = metadata.get('author_email', 'help@stsci.edu')
URL = metadata.get('url', 'https://github.com/spacetelescope/tweakwcs')
LICENSE = metadata.get('license', 'BSD-3-Clause')

# load long description
this_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_dir, LONG_DESCRIPTION), encoding='utf-8') as f:
    long_description = f.read()


def get_transforms_data():
    # Installs the schema files in jwst/transforms
    # Because the path to the schemas includes "stsci.edu" they
    # can't be installed using setuptools.
    transforms_schemas = []
    root = os.path.join(PACKAGENAME, 'schemas')

    if not os.path.isdir(root):
        return transforms_schemas

    for node, dirs, files in os.walk(root):
        for fname in files:
            if fname.endswith('.yaml'):
                transforms_schemas.append(
                    os.path.relpath(os.path.join(node, fname), root))
    # In the package directory, install to the subdirectory 'schemas'
    transforms_schemas = [os.path.join('schemas', s) for s in transforms_schemas]
    return transforms_schemas


PACKAGE_DATA = {
    '': [
        'README.rst',
        'LICENSE.txt',
        'CHANGELOG.rst',
        '*.fits',
        '*.txt',
        '*.inc',
        '*.cfg',
        '*.csv',
        '*.yaml',
        '*.json'
    ],
    'tweakwcs': get_transforms_data()
}

INSTALL_REQUIRES = [
    'numpy',
    'astropy>=3.1',
    'stsci.stimage',
    'stsci.imagestats',
    'spherical_geometry>=1.2.20',
]

TESTS_REQUIRE = [
    'pytest',
    'pytest-cov',
    'codecov',
    'scipy'
]

DOCS_REQUIRE = [
    'numpydoc',
    'sphinx',
    'sphinx_automodapi',
    'sphinx_rtd_theme',
    'stsci_rtd_theme',
    'graphviz',
]

setup(
    name=PACKAGENAME,
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    url=URL,
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Development Status :: 3 - Alpha',
    ],
    python_requires='>=3.5',
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    packages=find_packages(),
    package_data=PACKAGE_DATA,
    extras_require={
        'docs': DOCS_REQUIRE,
        'test': TESTS_REQUIRE,
    },
    project_urls={
        'Bug Reports': 'https://github.com/spacetelescope/tweakwcs/issues/',
        'Source': 'https://github.com/spacetelescope/tweakwcs/',
        'Help': 'https://hsthelp.stsci.edu/',
    },
)



from itertools import product

import numpy as np
from astropy.io import fits
from astropy import table
from astropy import wcs
from astropy.utils.data import get_pkg_data_filename
import pytest

import tweakwcs


def _match(x, y):
    lenx = len(x)
    leny = len(y)
    if lenx == leny:
        return (np.arange(lenx), np.arange(leny))
    elif lenx < leny:
        lenx, leny = leny, lenx
        x, y = y, x
    match = (np.arange(leny) + (0 if y.meta['name'] == 'ext1' else leny),
             np.arange(leny))
    return match


def test_multichip_fitswcs_alignment():
    h1 = fits.Header.fromfile(get_pkg_data_filename('data/wfc3_uvis1.hdr'))
    w1 = wcs.WCS(h1)
    imcat1 = tweakwcs.FITSWCS(w1)
    imcat1.meta['catalog'] = table.Table.read(
        get_pkg_data_filename('data/wfc3_uvis1.cat'),
        format='ascii.csv',
        delimiter=' ',
        names=['x', 'y']
    )
    imcat1.meta['group_id'] = 1
    imcat1.meta['name'] = 'ext1'

    h2 = fits.Header.fromfile(get_pkg_data_filename('data/wfc3_uvis2.hdr'))
    w2 = wcs.WCS(h2)
    imcat2 = tweakwcs.FITSWCS(w2)
    imcat2.meta['catalog'] = table.Table.read(
        get_pkg_data_filename('data/wfc3_uvis2.cat'),
        format='ascii.csv',
        delimiter=' ',
        names=['x', 'y']
    )
    imcat2.meta['group_id'] = 1
    imcat2.meta['name'] = 'ext4'

    refcat = table.Table.read(
        get_pkg_data_filename('data/ref.cat'),
        format='ascii.csv', delimiter=' ',
        names=['RA', 'DEC']
    )

    tweakwcs.align_wcs([imcat1, imcat2], refcat, match=_match, nclip=None,
                       sigma=3, fitgeom='general')

    fi1 = imcat1.meta['fit_info']
    fi2 = imcat2.meta['fit_info']

    w1 = imcat1.wcs
    w2 = imcat2.wcs

    assert np.allclose(w1.wcs.crval, (83.206917667519, -67.73275818507248), rtol=0)
    assert np.allclose(
        w1.wcs.cd,
        np.array(
            [
                [3.93222694902149e-06, -1.0106698270131359e-05],
                [-1.0377001075437075e-05, -4.577945148472431e-06]
            ]),
        atol=0.0,
        rtol=1e-8
    )

    assert np.allclose(w2.wcs.crval, (83.15167050722597, -67.74220306069903), rtol=0)
    assert np.allclose(
        w2.wcs.cd,
        np.array(
            [
                [3.834449806681195e-06, -9.996495217498745e-06],
                [-1.0348147451241423e-05, -4.503496019301529e-06]
            ]),
        atol=0.0,
        rtol=1e-8
    )

    assert np.allclose(fi1['<scale>'], 1.0025, rtol=0, atol=2e-8)
    assert np.allclose(fi2['<scale>'], 1.0025, rtol=0, atol=2e-8)

    assert fi1['rmse'] < 5e-5
    assert fi2['rmse'] < 5e-5

    cat1 = imcat1.meta['catalog']
    ra1, dec1 = w1.all_pix2world(cat1['x'], cat1['y'], 0)

    cat2 = imcat2.meta['catalog']
    ra2, dec2 = w2.all_pix2world(cat2['x'], cat2['y'], 0)

    ra = np.concatenate([ra1, ra2])
    dec = np.concatenate([dec1, dec2])

    rmse_ra = np.sqrt(np.mean((ra - refcat['RA'])**2))
    rmse_dec = np.sqrt(np.mean((dec - refcat['DEC'])**2))

    assert rmse_ra < 1e-9
    assert rmse_dec < 1e-9


@pytest.mark.parametrize('wcsno, refscale, dra, ddec', (
    x for x in product(
        [0, 1],
        [((15, 25), (1.0005, 0.9993))],
        [0, 0.05, -0.05, 4, -4],
        [0, 0.05, -0.05, 4, -4]
    )
))
def test_different_ref_tpwcs_fitswcs_alignment(wcsno, refscale, dra, ddec):
    # This test was designed to check that the results of alignment,
    # in particular and most importantly, the sky positions of sources in
    # aligned images do not depend on the tangent reference plane used
    # for alignment. [#125]
    h1 = fits.Header.fromfile(get_pkg_data_filename('data/wfc3_uvis1.hdr'))
    w1 = wcs.WCS(h1)
    imcat1 = tweakwcs.FITSWCS(w1)
    imcat1.meta['catalog'] = table.Table.read(
        get_pkg_data_filename('data/wfc3_uvis1.cat'),
        format='ascii.csv',
        delimiter=' ',
        names=['x', 'y']
    )
    imcat1.meta['group_id'] = 1
    imcat1.meta['name'] = 'ext1'

    h2 = fits.Header.fromfile(get_pkg_data_filename('data/wfc3_uvis2.hdr'))
    w2 = wcs.WCS(h2)
    imcat2 = tweakwcs.FITSWCS(w2)
    imcat2.meta['catalog'] = table.Table.read(
        get_pkg_data_filename('data/wfc3_uvis2.cat'),
        format='ascii.csv',
        delimiter=' ',
        names=['x', 'y']
    )
    imcat2.meta['group_id'] = 1
    imcat2.meta['name'] = 'ext4'

    refcat = table.Table.read(
        get_pkg_data_filename('data/ref.cat'),
        format='ascii.csv', delimiter=' ',
        names=['RA', 'DEC']
    )

    refwcses = [wcs.WCS(h1), wcs.WCS(h2)]

    refwcs = refwcses[wcsno]

    # change pointing of the reference WCS (alignment tangent plane):
    refwcs.wcs.crval = refwcses[1 - wcsno].wcs.crval + np.asarray([dra, ddec])

    rotm = tweakwcs.linearfit.build_fit_matrix(*refscale)
    refwcs.wcs.cd = np.dot(refwcs.wcs.cd, rotm)
    refwcs.wcs.set()
    ref_tpwcs = tweakwcs.FITSWCS(refwcs)

    tweakwcs.align_wcs([imcat1, imcat2], refcat, ref_tpwcs=ref_tpwcs,
                       match=_match, nclip=None, sigma=3, fitgeom='general')

    fi1 = imcat1.meta['fit_info']
    fi2 = imcat2.meta['fit_info']

    w1 = imcat1.wcs
    w2 = imcat2.wcs

    assert fi1['rmse'] < 1e-4
    assert fi2['rmse'] < 1e-4

    cat1 = imcat1.meta['catalog']
    ra1, dec1 = w1.all_pix2world(cat1['x'], cat1['y'], 0)

    cat2 = imcat2.meta['catalog']
    ra2, dec2 = w2.all_pix2world(cat2['x'], cat2['y'], 0)

    ra = np.concatenate([ra1, ra2])
    dec = np.concatenate([dec1, dec2])

    rmse_ra = np.sqrt(np.mean((ra - refcat['RA'])**2))
    rmse_dec = np.sqrt(np.mean((dec - refcat['DEC'])**2))

    assert rmse_ra < 5e-9
    assert rmse_dec < 5e-9
